dofile_once( "data/scripts/lib/utilities.lua" )
dofile_once( "data/scripts/invisibility_check.lua" )
local entity_id = GetUpdatedEntityID()
local x, y = EntityGetTransform(entity_id)

function split_string(inputstr, sep)
  sep = sep or "%s"
  local t= {}
  for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
    table.insert(t, str)
  end
  return t
end

function get_material_damage(entity_id)
  local damage_model_component = EntityGetFirstComponentIncludingDisabled(entity_id, "DamageModelComponent")
  if damage_model_component then
    local materials_that_damage = ComponentGetValue2(damage_model_component, "materials_that_damage") --"water,acid,blood,whatever"
    local materials_how_much_damage = ComponentGetValue2(damage_model_component, "materials_how_much_damage")--"1,-1,0.03,0.2"
    materials_that_damage = split_string(materials_that_damage, ",")
    materials_how_much_damage = split_string(materials_how_much_damage, ",")
    local material_damages = {}
    for i, v in ipairs(materials_that_damage) do
      material_damages[v] = tonumber(materials_how_much_damage[i])
    end
    return material_damages
  end
end

local entities = EntityGetInRadiusWithTag(x, y, 9999999, "mortal")

for i, entity in ipairs(entities) do

	local material_damages = get_material_damage(entity)

	if material_damages and material_damages.acid and material_damages.acid > 0 then

		local entity_damage_model_component = EntityGetFirstComponent(entity, "DamageModelComponent")
		
		if entity_damage_model_component then
		
			EntitySetDamageFromMaterial( entity, "cloud_acid", material_damages.acid )
			
		end
		
	end

end

function get_state()
  local entity_id = GetUpdatedEntityID()
  state = state or {}
  state[entity_id] = state[entity_id] or {}
  return state[entity_id]
end

--disable Ghost component if player is in holy mountain:

function get_state2()
  state2 = state2 or {}
  state2[entity_id] = state2[entity_id] or {
    last_frame_hit = 0
  }
  return state2[entity_id]
end

local state2 = get_state2()
local frames_active_after_getting_hit = 600

function damage_received()
  state2.last_frame_hit = GameGetFrameNum()
end

local is_angry = state2.last_frame_hit + frames_active_after_getting_hit > GameGetFrameNum()

--local x,y = EntityGetTransform( entity_id )
function get_distance(dx, dy, dx2, dy2)
	return math.sqrt(((dx-dx2)^2) + ((dy-dy2)^2))
end
--local x,y = EntityGetTransform( entity_id )
local players = {}
local player = EntityGetWithName( "DEBUG_NAME:player" )
if ( player ~= nil ) and ( player > 0 ) then
	players = {player}
else
	local polymorphed_player = EntityGetClosestWithTag(x, y, "polymorphed_player")
	if ( polymorphed_player ~= nil ) and ( polymorphed_player > 0 ) then
		players = {polymorphed_player}
	end
end

local ghost_component = EntityGetFirstComponentIncludingDisabled(entity_id, "GhostComponent")

if not is_angry and BiomeMapGetName() == "$biome_holymountain" and #players > 0 then
	EntitySetComponentIsEnabled( entity_id, ghost_component, false )
elseif is_angry and BiomeMapGetName() == "$biome_holymountain" and #players > 0 then
	EntitySetComponentIsEnabled( entity_id, ghost_component, true )
else
	EntitySetComponentIsEnabled( entity_id, ghost_component, true )
end

local times = ComponentGetValue2(GetUpdatedComponentID(), "mTimesExecuted")

if ( #players > 0 ) and ( is_invisible(players[1]) == true ) then

	for _, child in ipairs(EntityGetAllChildren(entity_id) or {} ) do
		if EntityGetName(child) == "Ghost Target" then
			target_id = child
			tx, ty = EntityGetTransform(target_id)
		end
	end
	
	SetRandomSeed( GameGetFrameNum(), x + y + entity_id )
	
	if times % 12 == 0 then
		EntitySetTransform( target_id, x + Random(-200, 200), y + Random(-200, 200))
		ComponentSetValue2( ghost_component, "mNextTargetCheckFrame", GameGetFrameNum() )
	end
	
	tx, ty = EntityGetTransform(target_id)
	
	local ghost_component = EntityGetFirstComponentIncludingDisabled(entity_id, "GhostComponent")
	if ( ghost_component ~= nil ) and ( ghost_component > 0 ) then
		ComponentSetValue2( ghost_component, "mTargetEntityId", target_id )
		ComponentSetValue2( ghost_component, "mTargetPosition", tx, ty )
		ComponentSetValue2( ghost_component, "mRandomTarget", tx, ty )
	end
end