function get_state()
  local entity_id = GetUpdatedEntityID()
  state = state or {}
  state[entity_id] = state[entity_id] or {}
  return state[entity_id]
end

local state = get_state()

function get_magnitude( x, y )
    local result = math.sqrt( x ^ 2 + y ^ 2 )
    return result
end

function vec_normalize(x, y)
    local m = get_magnitude(x, y)
    if m == 0 then return 0,0 end
    x = x / m
    y = y / m
    return x,y
end

state.dir = state.dir or 1
state.last_frame_bounced = state.last_frame_bounced or GameGetFrameNum()

local entity_id = GetUpdatedEntityID()
local x, y = EntityGetTransform(entity_id)
local vel_comp = EntityGetFirstComponent(entity_id, "VelocityComponent")
local vx, vy = ComponentGetValue2(vel_comp, "mVelocity")
if vy >= 0 and vy < 0.5 then
  vy = 1
end
vx, vy = vec_normalize(vx, vy)

local found_normal,normal_x,normal_y,approximate_distance_from_surface = GetSurfaceNormal(x+vx, y+vy, 5, 16)

local inside_wall, hit_x, hit_y = RaytraceSurfaces(x, y, x, y )

local found, nx, ny, apd = GetSurfaceNormal(x, y, 5, 16)
if found then
  nx = nx * (math.abs(apd) + 2)
  ny = ny * (math.abs(apd) + 2)
else
  nx = 0
  ny = 0
end

if inside_wall then

	EntitySetTransform(entity_id, x - nx, y - ny)

end

if found_normal then
  if vy >= 0 and state.last_frame_bounced < GameGetFrameNum() - 20 then
    ComponentSetValue2(vel_comp, "mVelocity", 100 * state.dir, -200)
    state.dir = state.dir * -1
    state.last_frame_bounced = GameGetFrameNum()
	
	EntityLoad("data/entities/particles/bouncer_particle.xml", hit_x, hit_y)
	GamePlaySound("data/audio/Desktop/animals.bank", "animals/spearbot/jump", hit_x, hit_y)
  end
end

function get_distance( x1, y1, x2, y2 )
	local result = math.sqrt( ( x2 - x1 ) ^ 2 + ( y2 - y1 ) ^ 2 )
	return result
end

local players = {}
local player = EntityGetWithName( "DEBUG_NAME:player" )
if ( player ~= nil ) and ( player > 0 ) then
	local px, py = EntityGetTransform(player)
	local dist = get_distance(x, y, px, py)
	if dist <= 300 then
		players = {player}
	end
else
	local polymorphed_player = EntityGetClosestWithTag(rx, ry, "polymorphed_player")
	if ( polymorphed_player ~= nil ) and ( polymorphed_player > 0 ) then
		local px, py = EntityGetTransform(polymorphed_player)
		local dist = get_distance(x, y, px, py)
		if dist <= 300 then
			players = {polymorphed_player}
		end
	end
end

-- if #players > 0 then

	-- ComponentSetValue2(GetUpdatedComponentID(), "execute_every_n_frame", 5)
	-- ComponentSetValue2(GetUpdatedComponentID(), "mNextExecutionTime", GameGetFrameNum())
-- else

	-- ComponentSetValue2(GetUpdatedComponentID(), "execute_every_n_frame", 200)

-- end