--Rotation
dofile_once( "data/scripts/invisibility_check.lua" )
dofile_once("data/scripts/lib/utilities.lua")

local entity_id = GetUpdatedEntityID()

local x, y, r = EntityGetTransform(entity_id)

local variable_storage = EntityGetComponent(entity_id, "VariableStorageComponent")

local aim_direction = {x = 0, y = 0}

local players = {}
local player = EntityGetWithName( "DEBUG_NAME:player" )
if ( player ~= nil ) and ( player > 0 ) then
	local px, py = EntityGetTransform(player)
	local dist = get_distance(x, y, px, py)
	if dist <= 300 then
		players = {player}
	end
else
	local polymorphed_player = EntityGetClosestWithTag(x, y, "polymorphed_player")
	if ( polymorphed_player ~= nil ) and ( polymorphed_player > 0 ) then
		local px, py = EntityGetTransform(polymorphed_player)
		local dist = get_distance(x, y, px, py)
		if dist <= 300 then
			players = {polymorphed_player}
		end
	end
end

if #players > 0 then

	local player_x, player_y = EntityGetTransform(players[1])

	local direction_x, direction_y = player_x - x, player_y - y

	local len = math.sqrt((direction_x*direction_x) + (direction_y*direction_y))

	aim_direction.x, aim_direction.y = direction_x / len, direction_y / len

	local phys_body = EntityGetFirstComponent(entity_id, "PhysicsBodyComponent")

	local ang_vel = PhysicsGetComponentAngularVelocity( entity_id, phys_body )

	function angle_rad(direction)
		return math.atan2(direction.y,direction.x)
	end

	function get_torque(current_angle, target_angle, speed)
		-- Make angle go from 0 to 2pi
		if target_angle < 0 then
		  target_angle = target_angle + math.pi * 2
		end
		if current_angle < 0 then
		  current_angle = current_angle + math.pi * 2
		end
	  
		local dist_up = math.abs(target_angle - current_angle)
		local dist_down = math.abs((current_angle + math.pi * 2) - target_angle)
		
		if current_angle > target_angle then
		  dist_up = math.abs((target_angle + math.pi * 2) - current_angle)
		  dist_down = math.abs(current_angle - target_angle)
		end
		
		local dir
		if math.abs(dist_up) < math.abs(dist_down) then
		  dir = 1
		else
		  dir = -1
		end
		local dist_smallest = dist_up
		if dist_down < dist_up then
		  dist_smallest = dist_down
		end
	  
		local new_angle = speed * dir

		return new_angle
	end

	PhysicsApplyTorque( entity_id, -ang_vel * 10 )

	PhysicsApplyTorque( entity_id, get_torque(r, angle_rad(aim_direction), 5) * 10 )
end

local number = GetValueNumber("number", 0)

for i, child in ipairs(EntityGetAllChildren(entity_id) or {} ) do
	if EntityGetName(child) == "Thun" then
		--Particle Effects:
		
		local times_executed = ComponentGetValue2(GetUpdatedComponentID(), "mTimesExecuted")

		if times_executed < 50 then
			EntitySetTransform(child, x, y)
		else
			EntitySetComponentsWithTagEnabled(child, "enabled_in_world", false)
		end

		local particle_emitter_component = EntityGetFirstComponent(child, "ParticleEmitterComponent")

		ComponentSetValue2(particle_emitter_component, "mExPosition", x, y)

		--Area Checker:

		local cx, cy = EntityGetTransform(child)
		local physics_body_component = EntityGetFirstComponentIncludingDisabled(child, "PhysicsBodyComponent")

		function get_distance( x1, y1, x2, y2 )
			local result = math.sqrt( ( x2 - x1 ) ^ 2 + ( y2 - y1 ) ^ 2 )
			return result
		end

		function get_direction( x1, y1, x2, y2 )
			return math.atan2( ( y2 - y1 ), ( x2 - x1 ) )
		end

		local distance = get_distance(cx, cy, x, y)
		local max_distance = 50
		local cx_new = cx
		local cy_new = cy
		if distance > max_distance then     
		  local direction = get_direction(x, y, cx, cy)
		  cx_new = x + math.cos(direction) * max_distance
		  cy_new = y + math.sin(direction) * max_distance
		  EntitySetComponentIsEnabled( child, physics_body_component, false )
		end

		EntitySetTransform(child, cx_new, cy_new)
		EntitySetComponentIsEnabled( child, physics_body_component, true )


		--Teleport to parent if inside wall:

		local did_hit, hit_x, hit_y = RaytracePlatforms(cx_new, cy_new, cx_new, cy_new)
		  
		if did_hit then

			-- EntityLoad( "data/entities/particles/teleportation_source.xml", x, y )
			-- EntityLoad( "data/entities/particles/teleportation_target.xml", parent_x, parent_y )
			EntitySetComponentIsEnabled( child, physics_body_component, false )
			EntitySetTransform( child, x, y )
			EntitySetComponentIsEnabled( child, physics_body_component, true )

		end
		
		--attack
		if ( times_executed % 100 == 0 ) and ( number <= 0 ) then

			if ( #players > 0 ) and ( EntityGetFirstComponent(players[1], "GenomeDataComponent") ~= nil ) and ( EntityGetFirstComponent(players[1], "GenomeDataComponent") > 0 ) and ( EntityGetHerdRelation(entity_id, players[1]) < 40 ) and ( is_invisible(players[1]) == false ) then
			  
				local player_x, player_y = EntityGetTransform(players[1])
				local did_hit, hit_x, hit_y = RaytracePlatforms(cx_new, cy_new, player_x, player_y)
			  
				if not did_hit then
					SetValueNumber("number", 160)
				end
			end
		end
		
		if number > 0 then
			
			if number == 160 then
				local child_id = EntityCreateNew()
				EntitySetTransform(child_id, cx_new, cy_new)
				EntityAddComponent(child_id, "InheritTransformComponent", {} )
				EntityAddChild(child, child_id)
				local comp = EntityAddComponent2(child_id, "ParticleEmitterComponent", {
					emitted_material_name="spark_green_bright",
					lifetime_min=0.5,
					lifetime_max=1.5,
					count_min=1,
					count_max=3,
					render_on_grid=true,
					fade_based_on_lifetime=true,
					cosmetic_force_create=false,
					airflow_force=0.5,
					airflow_time=0.01,
					airflow_scale=0.05,
					emission_interval_min_frames=1,
					emission_interval_max_frames=1,
					emit_cosmetic_particles=true,
					is_emitting=true,
					draw_as_long=true,
					attractor_force=100,
				} )
				ComponentSetValue2(comp, "gravity", 0, 0)
				ComponentSetValue2(comp, "area_circle_radius", 0, 20)
				EntityAddComponent(child_id, "AudioComponent", {
					file="data/audio/Desktop/projectiles.bank",
					event_root="player_projectiles/megalaser",
				} )

				EntityAddComponent(child_id, "LifetimeComponent", {
					lifetime="60",
				} )
					
			elseif number == 100 then
					
				local child_id2 = EntityCreateNew()
				EntitySetTransform(child_id2, cx_new, cy_new)
				EntityAddComponent(child_id2, "InheritTransformComponent", {} )
				EntityAddChild(child, child_id2)
				local comp2 = EntityAddComponent2(child_id2, "ParticleEmitterComponent", {
					emitted_material_name="spark_green_bright",
					x_pos_offset_min=0,
					y_pos_offset_min=0,
					x_pos_offset_max=0,
					y_pos_offset_max=0,
					x_vel_min=0,
					x_vel_max=0,
					y_vel_min=0,
					y_vel_max=0,
					friction=4,
					count_min=30,
					count_max=50,
					velocity_always_away_from_center=80,
					lifetime_min=0.9,
					lifetime_max=3.2,
					render_on_grid=false,
					airflow_force=0.25,
					airflow_time=0.101,
					airflow_scale=2.01,
					fade_based_on_lifetime=true,
					create_real_particles=false,
					emit_cosmetic_particles=true,
					emission_interval_min_frames=5,
					emission_interval_max_frames=5,
					delay_frames=8,
					is_emitting=true,
				} )
				ComponentSetValue2(comp2, "gravity", 0, 0)
				ComponentSetValue2(comp2, "offset", 0, 0)
				ComponentSetValue2(comp2, "area_circle_radius", 4, 4)
				EntityAddComponent(child_id2, "LuaComponent", {
					script_source_file="data/scripts/projectiles/thou_laser_projectile_spawner.lua",
					execute_every_n_frame="5",
				} )
				EntityAddComponent(child_id2, "LifetimeComponent", {
					lifetime="20",
				} )
			end
		end
	end
end

if number > 0 then
	SetValueNumber("number", number - 1)
end

if #players > 0 then
	ComponentSetValue2(GetUpdatedComponentID(), "execute_every_n_frame", 1)
	EntitySetComponentsWithTagEnabled(child, "enabled_in_world", false)
else
	ComponentSetValue2(GetUpdatedComponentID(), "execute_every_n_frame", 200)
	EntitySetComponentsWithTagEnabled(child, "enabled_in_world", true)
end
