dofile_once("data/scripts/lib/utilities.lua")
dofile_once( "data/scripts/invisibility_check.lua" )

local root_id = GetUpdatedEntityID()
local rx, ry = EntityGetTransform( root_id )

local this_lua = GetUpdatedComponentID()

local players = {}
local player = EntityGetWithName( "DEBUG_NAME:player" )
if ( player ~= nil ) and ( player > 0 ) then
	players = {player}
else
	local polymorphed_player = EntityGetClosestWithTag(rx, ry, "polymorphed_player")
	if ( polymorphed_player ~= nil ) and ( polymorphed_player > 0 ) then
		players = {polymorphed_player}
	end
end

if #players > 0 then
	local player_x2, player_y2 = EntityGetTransform(players[1])
	local dist2 = get_distance(rx, ry, player_x2, player_y2)

	if dist2 < 300 then
		ComponentSetValue2(this_lua, "execute_every_n_frame", 2)
		ComponentSetValue2(this_lua, "mNextExecutionTime", GameGetFrameNum())
		for i, child in ipairs(EntityGetAllChildren(root_id) or {} ) do
			if EntityGetName(child) == "statue_hand" then
				EntitySetComponentsWithTagEnabled(child, "enabled_in_world", false)
			end
		end
	else
		ComponentSetValue2(this_lua, "execute_every_n_frame", 200)
		for i, child in ipairs(EntityGetAllChildren(root_id) or {} ) do
			if EntityGetName(child) == "statue_hand" then
				EntitySetComponentsWithTagEnabled(child, "enabled_in_world", true)
			end
		end
	end
	if dist2 > 600 then
		SetValueNumber("number", 0)
	end
end

local times_executed = ComponentGetValue2(this_lua, "mTimesExecuted")
local number = GetValueNumber("number", 0)

if number > 0 then
	SetValueNumber("number", number - 1)
end

for i, child in ipairs(EntityGetAllChildren(root_id) or {} ) do
	if EntityGetName(child) == "statue_hand" then
		
		local lerp_amount = 0.9
		local offset_x = 15
		local bob_h = 8
		local bob_w = 20
		local bob_speed_y = 0.023
		local bob_speed_x = 0.01121
		local rotate_amount = math.pi * 0.05
		local rotate_speed = 0.093

		local pos_x, pos_y = EntityGetTransform( child )

		-- handedness from variable storage
		local comp = get_variable_storage_component(child, "is_right")
		is_right = ComponentGetValue2(comp, "value_bool")
		if not is_right then offset_x = -offset_x end

		target_x = rx + offset_x

		local time = GameGetFrameNum()
		local r = ProceduralRandomf(entity_id, 0, -1, 1)

		-- randomize times and speeds slightly so that multiple fists don't fly identically
		time = time + r * 10000
		bob_speed_y = bob_speed_y + (r * bob_speed_y * 0.1)
		bob_speed_x = bob_speed_x + (r * bob_speed_x * 0.1)
		lerp_amount = lerp_amount - (r * lerp_amount * 0.01)

		-- bob
		target_y = ry + math.sin(time * bob_speed_y) * bob_h
		target_x = target_x + math.sin(time * bob_speed_x) * bob_w

		-- rotate
		rotate_amount = rotate_amount * math.sin(time * rotate_speed)

		-- move towards target
		pos_x, pos_y = vec_lerp(pos_x, pos_y, target_x, target_y, lerp_amount)
		EntitySetTransform( child, pos_x, pos_y, rotate_amount, 1, 1)


		
		--attack
		function spawn_portal(which_one, material)
			
			-- local opts = { "portal_orb_blue", "portal_orb_green", "portal_orb_pink" }
			
			SetRandomSeed( GameGetFrameNum(), pos_x + child )
			
			local raycasts = 4
			local dir = ( math.pi * 2.0 ) / raycasts
			local length = 8
			
			-- local rnd = Random( 1, #opts )
			local portal = "data/entities/projectiles/" .. which_one .. ".xml"
			
			if ( #players > 0 ) then
				local t = players[1]
				local px, py = EntityGetTransform( t )
				
				local dx = px + Random( -150, 150 )
				local dy = py + Random( -150, 150 )
				
				local dist = math.abs( py - dy ) + math.abs( px - dx )
				local limit = 0
				local wall = true
				
				while ( ( dist < 100 ) or wall ) and ( limit < 20 ) do
					dx = px + Random( -100, 100 )
					dy = py + Random( -100, 100 )
					dist = math.abs( py - dy ) + math.abs( px - dx )
					limit = limit + 1
					
					for j=0,raycasts-1 do
						local ex = dx + math.cos( j * dir ) * length
						local ey = dy - math.sin( j * dir ) * length
						
						wall = RaytraceSurfaces( dx, dy, ex, ey )
						
						if wall then
							break
						end
					end
				end
				if portal == "data/entities/projectiles/portal_orb_blue" then
					EntityLoad( "data/entities/particles/poof_blue.xml", dx, dy )
				end
				if portal == "data/entities/projectiles/portal_orb_green" then
					EntityLoad( "data/entities/particles/poof_blue.xml", dx, dy )
				end
				if portal == "data/entities/projectiles/portal_orb_pink" then
					EntityLoad( "data/entities/particles/poof_pink.xml", dx, dy )
				end
				if portal == "data/entities/projectiles/portal_orb_purple" then
					EntityLoad( "data/entities/particles/poof_pink.xml", dx, dy )
					EntityLoad( "data/entities/particles/poof_blue.xml", dx, dy )
				end
				local portal_id = EntityLoad( portal, dx, dy )
				EntityAddChild(child, portal_id)
				
				-- Add Trail Particles
					
				local particle_comp = EntityAddComponent2(portal_id, "ParticleEmitterComponent", {
						  emitted_material_name=material,
						  x_pos_offset_min=-2,
						  x_pos_offset_max=2,
						  y_pos_offset_min=-2,
						  y_pos_offset_max=2,
						  x_vel_min=-50,
						  x_vel_max=50,
						  y_vel_min=-50,
						  y_vel_max=50,
						  count_min=0,
						  count_max=1,
						  is_trail=true,
						  trail_gap=1,
						  lifetime_min=0.8,
						  lifetime_max=0.8,
						  fade_based_on_lifetime=true,
						  create_real_particles=false,
						  emit_cosmetic_particles=true,
						  collide_with_grid=false,
						  emission_interval_min_frames=0,
						  emission_interval_max_frames=0,
						  airflow_force=0.36,
						  airflow_time=1.20,
						  airflow_scale=0.28,
						  is_emitting=true,
				} )
				ComponentSetValue2(particle_comp, "gravity", 0, 0)
				ComponentSetValue2(particle_comp, "offset", 0, 0)
				
				EntityAddComponent(portal_id, "LuaComponent", {
					script_source_file="data/scripts/animals/head_statue_particles.lua",
					execute_every_n_frame="1",
					execute_on_added="1",
				} )
					
				-- EntityAddComponent(particles_id, "LifetimeComponent", {
					-- lifetime="240",
				-- } )
			end
		end
	
		if ( times_executed % 60 == 0 ) and ( number <= 0 ) then

			if #players > 0 and ( EntityGetFirstComponent(players[1], "GenomeDataComponent") ~= nil ) and ( EntityGetFirstComponent(players[1], "GenomeDataComponent") > 0 ) and ( EntityGetHerdRelation(root_id, players[1]) < 40 ) and ( is_invisible(players[1]) == false ) then
			  
				local player_x, player_y = EntityGetTransform(players[1])
				local did_hit, hit_x, hit_y = RaytraceSurfaces(pos_x, pos_y, player_x, player_y)
				local dist = get_distance(pos_x, pos_y, player_x, player_y)
			  
				if not did_hit and ( dist <= 250 ) then
					SetValueNumber("number", 600)
				end
			end
		end

		if number > 0 then
			
			if number == 600 then
			
				--Opening finger 1:
			
				edit_component2( child, "SpriteComponent", function(sprite_comp,vars)
						
					ComponentSetValue2( sprite_comp, "rect_animation", "open1")
						
				end)
				
				--Adding Vortex particle Effect:
				
				local vortex_id = EntityLoad( "data/particles/vortex/field1.xml", pos_x, pos_y )
				EntitySetTransform(vortex_id, pos_x, pos_y)
				EntityAddChild(child, vortex_id)
					
			elseif number == 570 then
					
				spawn_portal("portal_orb_green","spark_green")
				
			elseif number == 450 then
			
				edit_component2( child, "SpriteComponent", function(sprite_comp,vars)
						
					ComponentSetValue2( sprite_comp, "rect_animation", "open2")
						
				end)
				
				--Adding Vortex particle Effect:
				
				local vortex_id = EntityLoad( "data/particles/vortex/field3.xml", pos_x, pos_y )
				EntitySetTransform(vortex_id, pos_x, pos_y)
				EntityAddChild(child, vortex_id)
			elseif number == 420 then
				spawn_portal("portal_orb_pink","spark_red")
			elseif number == 300 then
				edit_component2( child, "SpriteComponent", function(sprite_comp,vars)
						
					ComponentSetValue2( sprite_comp, "rect_animation", "open3")
						
				end)
				
				--Adding Vortex particle Effect:
				
				local vortex_id = EntityLoad( "data/particles/vortex/field2.xml", pos_x, pos_y )
				EntitySetTransform(vortex_id, pos_x, pos_y)
				EntityAddChild(child, vortex_id)
			elseif number == 270 then
				spawn_portal("portal_orb_blue","spark_blue")
			elseif number == 150 then
				edit_component2( child, "SpriteComponent", function(sprite_comp,vars)
						
					ComponentSetValue2( sprite_comp, "rect_animation", "open4")
						
				end)
				
				--Adding Vortex particle Effect:
				
				local vortex_id = EntityLoad( "data/particles/vortex/field4.xml", pos_x, pos_y )
				EntitySetTransform(vortex_id, pos_x, pos_y)
				EntityAddChild(child, vortex_id)
			elseif number == 120 then
				spawn_portal("portal_orb_purple","spark_purple")
			elseif number == 1 then
				edit_component2( child, "SpriteComponent", function(sprite_comp,vars)
						
					ComponentSetValue2( sprite_comp, "rect_animation", "close")
						
				end)
			end
		end
	end
end