--dofile_once( "data/scripts/lib/utilities.lua" )

function get_distance(dx, dy, dx2, dy2)
    return math.sqrt(((dx-dx2)^2) + ((dy-dy2)^2))
end

function GetSurfaceNormal(pos_x, pos_y, ray_length, ray_count)
	local total_normal_x = 0
	local total_normal_y = 0
	local total_distance = 0
	local found_normal = false

	for i = 1, ray_count do
		local angle = (i - 1) * (2 * math.pi / ray_count)
		local ray_x = pos_x + ray_length * math.cos(angle)
		local ray_y = pos_y + ray_length * math.sin(angle)

		-- local did_hit_liquid_and_surfaces, hit_x2, hit_y2 = RaytraceSurfacesAndLiquiform(pos_x, pos_y, ray_x, ray_y)
		local did_hit_surfaces_only, hit_x, hit_y = RaytraceSurfaces(pos_x, pos_y, ray_x, ray_y)
		-- local in_liquid = false
		-- if did_hit_liquid_and_surfaces and ( not did_hit_surfaces_only ) then
			-- in_liquid = true
		-- end

		if did_hit_surfaces_only then
			found_normal = true
			local normal_x, normal_y = hit_x - pos_x, hit_y - pos_y
			total_normal_x = total_normal_x + normal_x
			total_normal_y = total_normal_y + normal_y
			total_distance = total_distance + math.sqrt(normal_x^2 + normal_y^2)
		end
	end

	if found_normal then
		local average_normal_x = total_normal_x / ray_count
		local average_normal_y = total_normal_y / ray_count
		local average_distance = total_distance / ray_count
		return true, average_normal_x, average_normal_y, average_distance
	else
		return false, 0, 0, 0
	end
end

function get_direction( x1, y1, x2, y2 )
  return math.atan2( ( y2 - y1 ), ( x2 - x1 ) )
end

local times = ComponentGetValue2(GetUpdatedComponentID(), "mTimesExecuted")
local frames_to_bite = 10

local entity_id = GetUpdatedEntityID()
local x, y, rot = EntityGetTransform(entity_id)

SetRandomSeed( GameGetFrameNum(), y + x + entity_id )

local components = EntityGetComponent( entity_id, "VariableStorageComponent" )
if ( components ~= nil ) then
	for _,comp_id in pairs(components) do 
		local var_name = ComponentGetValue( comp_id, "name" )
		if ( var_name == "phase_value") then
			var_comp_phase = comp_id
		end
		if ( var_name == "prevx") then
			var_comp_prevx = comp_id
		end
		if ( var_name == "prevy") then
			var_comp_prevy = comp_id
		end
		if ( var_name == "in_air_time") then
			var_comp_in_air_time = comp_id
		end
		if ( var_name == "below_air_time") then
			var_comp_below_air_time = comp_id
		end
		if ( var_name == "cooldown") then
			var_comp_cooldown = comp_id
		end
		if ( var_name == "lava_is_shifted_check") then
			var_comp_lava_is_shifted_check = comp_id
		end
		if ( var_name == "rock_is_shifted_check") then
			var_comp_rock_is_shifted_check = comp_id
		end
		if ( var_name == "lava_circle_entity") then
			var_comp_lava_circle_entity = comp_id
		end
		
		-- if ( var_name == "orbit_radius_max") then
			-- var_comp_orbit_radius_max = comp_id
		-- end
		-- if ( var_name == "orbit_radius") then
			-- var_comp_orbit_radius = comp_id
		-- end
	end
end

local state = ComponentGetValue2( var_comp_phase, "value_int" )

local prevx = ComponentGetValue2( var_comp_prevx, "value_float" )
local prevy = ComponentGetValue2( var_comp_prevy, "value_float" )

ComponentSetValue2(var_comp_prevx, "value_float", x)
ComponentSetValue2(var_comp_prevy, "value_float", y)

function calculateVelocity(prev_x, prev_y, x, y)
    -- Calculate the displacement in x and y directions
    local dx = x - prev_x
    local dy = y - prev_y
    
    -- Calculate velocity components (vx, vy)
    local vx = dx
    local vy = dy
    
    return vx, vy
end

function split_string(inputstr, sep)
  sep = sep or "%s"
  local t= {}
  for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
	table.insert(t, str)
  end
  return t
end

function get_material_damage(entity_id)
  local damage_model_component = EntityGetFirstComponentIncludingDisabled(entity_id, "DamageModelComponent")
  if damage_model_component then
	local materials_that_damage = ComponentGetValue2(damage_model_component, "materials_that_damage") --"water,acid,blood,whatever"
	local materials_how_much_damage = ComponentGetValue2(damage_model_component, "materials_how_much_damage")--"1,-1,0.03,0.2"
	materials_that_damage = split_string(materials_that_damage, ",")
	materials_how_much_damage = split_string(materials_how_much_damage, ",")
	local material_damages = {}
	for i, v in ipairs(materials_that_damage) do
	  material_damages[v] = tonumber(materials_how_much_damage[i])
	end
	return material_damages
  end
end

function get_fungal_shifted_material_name(material_name, multiple_materials_boolean)
	-- Get the shifts if not provided
	local world_state_entity_id = GameGetWorldStateEntity()
	local world_state_component = EntityGetFirstComponentIncludingDisabled(world_state_entity_id, "WorldStateComponent")
	local shifts = ComponentGetValue2(world_state_component, "changed_materials")

	-- Function to find shifts for a single material
	local function find_shifts(target)
		local results = {}
		for i = #shifts - 1, 1, -2 do
			local from = shifts[i]
			local to = shifts[i + 1]

			if target == from or target == to then
				table.insert(results, {from, to})
			end
		end
		return results
	end

	if multiple_materials_boolean and type(material_name) == "table" then
		-- Handle multiple materials
		local all_results = {}
		for _, name in ipairs(material_name) do
			local results = find_shifts(name)
			if #results > 0 then
				all_results[name] = results
			end
		end
		return all_results
	else
		-- Handle single material
		local results = find_shifts(material_name)
		if #results > 0 then
			return results[1][1], results[1][2]
		end
	end
end

if ComponentGetValue2( var_comp_lava_is_shifted_check, "value_int" ) == 0 then
	if times % 200 == 0 then
		local from, to = get_fungal_shifted_material_name("lava")
		
		if from == "lava" then
			local components = EntityGetComponentIncludingDisabled(entity_id, "MagicConvertMaterialComponent")
			if components then
				for k, component in pairs(components)do
					ComponentSetValue2(component, "to_material", CellFactory_GetType( "new_enemies_custom_lava" ))
					ComponentSetValue2(component, "to_material_array", "new_enemies_custom_lava,new_enemies_custom_lava,new_enemies_custom_lava,new_enemies_custom_lava,new_enemies_custom_lava")
				end
			end
			ComponentSetValue2(var_comp_lava_circle_entity, "value_string", "data/entities/projectiles/deck/lava_circle_custom2.xml")
			ComponentSetValue2(var_comp_lava_is_shifted_check, "value_int", 1)
		end
	end
end

if ComponentGetValue2(var_comp_rock_is_shifted_check, "value_int") == 0 then
	if times % 200 == 0 then
		local materials = {
			"rock_static",
			"lavarock_static",
			"templebrick_golden_static",
			"rock_static_radioactive",
			"rock_static_poison"
		}
		local shifted_materials = get_fungal_shifted_material_name(materials, true)

		for _, name in ipairs(materials) do
			local shifts = shifted_materials[name]
			if shifts then
				for _, shift in ipairs(shifts) do
					local from = shift[1]
					local to = shift[2]
					if from == name or to == name then
						EntitySetComponentsWithTagEnabled(entity_id, "enabled_in_world", true)
					end
				end
			end
		end
	end
end

local worm_component = EntityGetFirstComponentIncludingDisabled( entity_id, "WormComponent")
-- if ( worm_component ~= nil ) and ( worm_component > 0 ) then
	-- vx, vy = ComponentGetValue2(worm_component, "mTargetVec")
-- end

local vx, vy = calculateVelocity(prevx, prevy, x, y)

local target_kill_radius = ComponentGetValue2(worm_component, "target_kill_radius")

local child_entities = EntityGetAllChildren(entity_id) or {}
local face_entity
for i, v in ipairs(child_entities) do
  if EntityGetName(v) == "face_worm_head" then
    face_entity = v
	head_grab_var_component = EntityGetFirstComponentIncludingDisabled( face_entity, "VariableStorageComponent")
    break
  end
end

-- local vx, vy = GameGetVelocityCompVelocity( face_entity )
-- GamePrint(tostring(vx) .. " " .. tostring(vy))

local entities = EntityGetInRadiusWithTag(x + vx * frames_to_bite, y + vy * frames_to_bite, target_kill_radius + 30, "hittable")


local sprite_component = EntityGetFirstComponent(face_entity, "SpriteComponent")

local fx, fy, frot = EntityGetTransform(face_entity)

if ( ( vx < 0.5 ) and ( vx > -0.5 ) ) and ( ( vy < 0.5 ) and ( vy > -0.5 ) ) then
	EntitySetTransform(face_entity, fx, fy, frot)
else
	local dir = get_direction(x, y, x - vx, y - vy)
	-- GamePrint(tostring(vx) .. " " .. tostring(vy))
	EntitySetTransform(face_entity, fx, fy, dir)
end

--Get Jaw Position
-- local mandible = { x1 = 150, y1 = 0 }
-- local mandible_x, mandible_y = vec_rotate(mandible.x1, mandible.y1, dir)
-- mandible_x = x + mandible_x
-- mandible_y = y + mandible_y

local mFrameNextDamage = ComponentGetValue2(worm_component, "mFrameNextDamage")

local is_grabbing = ComponentGetValue2(head_grab_var_component, "value_bool")

if #entities > 1 and (GameGetFrameNum() - mFrameNextDamage > 0) and ( is_grabbing == false ) then
  local an = ComponentGetValue2(sprite_component, "rect_animation")
  if an ~= "bite" then
    ComponentSetValue2(sprite_component, "rect_animation", "bite")
  end
end

--Launch out of water and revert
function is_in_liquid_checker(x2, y2)
	local did_hit_liquid_and_surfaces, hit_x2, hit_y2 = RaytraceSurfacesAndLiquiform(x2, y2, x2, y2)
	local did_hit_surfaces_only, hit_x, hit_y = RaytraceSurfaces(x2, y2, x2, y2)
	if not ( did_hit_liquid_and_surfaces and ( not did_hit_surfaces_only ) ) then
		return false
	else
		return true
	end
end

function calculate_orbit_positions(center_x, center_y, radius, angle_offset, num_buds)
    local positions = {}
    for i = 0, num_buds - 1 do
        local angle = (i / num_buds) * 2 * math.pi + angle_offset
        local x2 = center_x + radius * math.cos(angle)
        local y2 = center_y + radius * math.sin(angle)
        table.insert(positions, {x = x2, y = y2, angle = angle})
    end
    return positions
end

function distance(x1,y1,x2,y2)
    return math.sqrt((x1 - x2)^2 + (y1 - y2)^2)
end

function splash(x1,y1)
	-- Define the radius of the area
	local radius = 100  -- You can adjust the radius as needed

	-- Iterate through each pixel in the area
	for dx = -radius, radius, 5 do
		for dy = -radius, radius, 5 do
			local px = x1 + dx
			local py = y1 + dy
			-- Check if the pixel is within the radius
			if (dx * dx + dy * dy) <= radius * radius then
				-- Spawn the entity at the current pixel
				local splash_entity = EntityLoad("mods/new_enemies/files/liquid_splash/liquid_splash.xml", px, py)
				local dist_from_center = distance(x,y,px,py)
				local MaterialInventoryComponent = EntityGetFirstComponent(splash_entity, "MaterialInventoryComponent")

				local factor = 2 - math.min(1, dist_from_center / radius) * 2
				-- local steepness_factor = 2
				-- factor = math.pow(factor, steepness_factor)
				ComponentSetValue2(MaterialInventoryComponent, "death_throw_particle_velocity_coeff", factor)
				
			end
		end
	end
	-- GamePlaySound( "data/audio/Desktop/materials.bank", "materials/liquid_displaced", x1, y1 )
	-- GamePlaySound( "data/audio/Desktop/materials.bank", "materials/liquid_splash", x1, y1 )
	-- GamePlaySound( "data/audio/Desktop/materials.bank", "materials/liquid_splash_player", x, y )
	--GamePlaySound( "data/audio/Desktop/player.bank", "player/jump_water", x1, y1 )
	-- GamePlaySound( "data/audio/Desktop/materials.bank", "materials/liquid_splash_player", x1, y1 )
	-- GamePlaySound( "data/audio/Desktop/materials.bank", "materials/liquid_splash", x1, y1 )
end

-- GamePlaySound( "data/audio/Desktop/player.bank", "player/land_water", x, y )

--GOTO PLAYER
local players = {}
local player = EntityGetWithName( "DEBUG_NAME:player" )
if ( player ~= nil ) and ( player > 0 ) then
	local px, py = EntityGetTransform(player)
	local dist = get_distance(x, y, px, py)
	if dist <= 999 then
		players = {player}
	end
else
	local polymorphed_player = EntityGetClosestWithTag(x, y, "polymorphed_player")
	if ( polymorphed_player ~= nil ) and ( polymorphed_player > 0 ) then
		local px, py = EntityGetTransform(polymorphed_player)
		local dist = get_distance(x, y, px, py)
		if dist <= 999 then
			players = {polymorphed_player}
		end
	end
end

local time_exposed = ComponentGetValue2( var_comp_in_air_time, "value_int" )
local time_below_air = ComponentGetValue2( var_comp_below_air_time, "value_int" )

local lava_circle_entity_string = ComponentGetValue2( var_comp_lava_circle_entity, "value_string" )

if state <= 1 then
	if ( time_exposed > 50 ) then
		EntityLoad(lava_circle_entity_string, x, y)
		ComponentSetValue2(var_comp_in_air_time, "value_int", 0)
	end

	if ( time_below_air > 50 ) and ( time_exposed <= 0 ) then
		EntityLoad(lava_circle_entity_string, x, y)
		ComponentSetValue2(var_comp_below_air_time, "value_int", 0)
	end

	if times % 12 == 0 then
		if (is_in_liquid_checker(x, y) == true ) and (is_in_liquid_checker(x, y - 40) == false) then
			ComponentSetValue2(var_comp_below_air_time, "value_int", time_below_air + 1)
		end
	end
	
	if is_grabbing == true then
		local tvx, tvy = ComponentGetValue2( worm_component, "mTargetVec" )
		ComponentSetValue2( worm_component, "mTargetVec", tvx, 1 )
	end
end

local cooldown = ComponentGetValue2( var_comp_cooldown, "value_int" )

local bud_entities = {}

local neck_entities = {}

local num_buds = 0

if #players > 0 then
	local px, py = EntityGetTransform(players[1])
	local worm_ai_component = EntityGetFirstComponentIncludingDisabled( entity_id, "WormAIComponent")
	-- local worm_component = EntityGetFirstComponentIncludingDisabled( entity_id, "WormComponent")
	if ( worm_ai_component ~= nil ) and ( worm_ai_component > 0 ) then
		ComponentSetValue2( worm_ai_component, "mTargetEntityId", players[1] )
	end
	if state <= 1 then
		if times % 12 == 0 then

			if ( worm_component ~= nil ) and ( worm_component > 0 ) then
				-- if is_in_liquid_checker(x, y) == false then
					-- ComponentSetValue2( worm_component, "is_water_worm", false )
					-- local current_speed = ComponentGetValue2(worm_component, "mSpeed")
					-- if current_speed > 0 then
						-- ComponentSetValue2( worm_component, "mSpeed", ComponentGetValue2(worm_component, "mSpeed") * 0.99 )
					-- end
					-- ComponentSetValue2( worm_component, "gravity", 1 )
					-- ComponentSetValue2( worm_component, "tail_gravity", 2 )
				-- elseif ( is_in_liquid_checker(x, y) == true ) and ( is_in_liquid_checker(mandible_x, mandible_y) == false ) then
					-- ComponentSetValue2( worm_component, "is_water_worm", false )
					-- ComponentSetValue2( worm_component, "mSpeed", 10 )
					-- ComponentSetValue2( worm_component, "gravity", 0 )
					-- ComponentSetValue2( worm_component, "tail_gravity", 0 )
				if ( is_in_liquid_checker(px, py) == true ) and ( is_in_liquid_checker(x, y) == true ) then
					ComponentSetValue2( worm_component, "is_water_worm", true )
					ComponentSetValue2( worm_component, "gravity", 0 )
					ComponentSetValue2( worm_component, "tail_gravity", 0 )
					ComponentSetValue2( worm_component, "mSpeed", 2 )
				elseif ( is_in_liquid_checker(px, py) == false ) and ( is_in_liquid_checker(x, y) == true ) and ( cooldown <= 0 ) then
					if times % 50 == 0 then
						ComponentSetValue2(var_comp_cooldown, "value_int", 8)
					end
				end
				if is_in_liquid_checker(x, y) == false then
					local current_speed = ComponentGetValue2(worm_component, "mSpeed")
					if current_speed > 0 then
						ComponentSetValue2( worm_component, "mSpeed", current_speed * 0.99 )
					end
					ComponentSetValue2( worm_component, "is_water_worm", false )
					ComponentSetValue2( worm_component, "gravity", 3 )
					ComponentSetValue2( worm_component, "tail_gravity", 4 )
					ComponentSetValue2(var_comp_in_air_time, "value_int", time_exposed + 1)
					if time_exposed == 0 then
						splash(x,y)
						GamePlaySound("mods/new_enemies/data/sound/more_sounds/my_mod_audio.bank", "snd_mod/splash_out", x, y)
					end
				else
					ComponentSetValue2(var_comp_in_air_time, "value_int", 0)
					if time_exposed > 0 then
						splash(x,y)
						GamePlaySound("mods/new_enemies/data/sound/more_sounds/my_mod_audio.bank", "snd_mod/splash", x, y)
					end
				end
			end
		end
		
		x, y, rot = EntityGetTransform(entity_id)
		
		local did_hit, hit_x, hit_y = RaytraceSurfaces(x, y, px, py)
		local far_away = distance(x,y,px,py)
		if ( did_hit or ( far_away > 300 ) ) and ( times % 100 == 0 ) then
			local ent = EntityLoad("data/entities/animals/hellion_small.xml", x, y )
			local child = EntityLoad( "data/entities/particles/hellion_absorb_particle.xml", x, y )
			EntitySetTransform(child, x, y)
			EntityAddChild(entity_id, child)
			
		  local an = ComponentGetValue2(sprite_component, "rect_animation")
		  if an ~= "bite" then
			ComponentSetValue2(sprite_component, "rect_animation", "bite")
		  end
			-- GamePlaySound( "data/audio/Desktop/projectiles.bank", "player_projectiles/black_hole_giga", x, y )
		elseif ( times % 600 == 0 ) and ( times > 1 ) then
			local ent = EntityLoad("data/entities/animals/hellion_small.xml", x, y )
			-- GamePlaySound( "data/audio/Desktop/projectiles.bank", "player_projectiles/black_hole_giga", x, y )
			local child = EntityLoad( "data/entities/particles/hellion_absorb_particle.xml", x, y )
			EntitySetTransform(child, x, y)
			EntityAddChild(entity_id, child)
			
		  local an = ComponentGetValue2(sprite_component, "rect_animation")
		  if an ~= "bite" then
			ComponentSetValue2(sprite_component, "rect_animation", "bite")
		  end
		end
			
		if cooldown > 0 then
			ComponentSetValue2(var_comp_cooldown, "value_int", cooldown - 1)
			-- local did_hit, hit_x, hit_y = RaytraceSurfaces(x, y, px, py)
			-- if not did_hit and cooldown > 100 then
				ComponentSetValue2( worm_component, "is_water_worm", false )
				ComponentSetValue2( worm_component, "mSpeed", 10 )
				ComponentSetValue2( worm_component, "gravity", 0 )
				ComponentSetValue2( worm_component, "tail_gravity", 0 )
			-- end
		end
	elseif state == 2 then
		local players_close = {}
		local player = EntityGetWithName( "DEBUG_NAME:player" )
		if ( player ~= nil ) and ( player > 0 ) then
			local px, py = EntityGetTransform(player)
			local dist = get_distance(x, y, px, py)
			if dist <= 512 then
				players_close = {player}
			end
		else
			local polymorphed_player = EntityGetClosestWithTag(x, y, "polymorphed_player")
			if ( polymorphed_player ~= nil ) and ( polymorphed_player > 0 ) then
				local px, py = EntityGetTransform(polymorphed_player)
				local dist = get_distance(x, y, px, py)
				if dist <= 512 then
					players_close = {polymorphed_player}
				end
			end
		end

		if #players_close > 0 then
		
			for i, child in ipairs(EntityGetAllChildren(entity_id) or {}) do
				if EntityGetName(child) == "flower_head" then
					EntitySetComponentsWithTagEnabled(child, "enabled_in_world", false)
					torso_id = child
					tx, ty, trot = EntityGetTransform(torso_id)
					
					table.insert(bud_entities, torso_id)
					
					num_buds = num_buds + 1
					
					-- local components = EntityGetComponent(torso_id, "MagicConvertMaterialComponent")
					-- if components then
						-- for k, component in pairs(components)do
							-- local old_radius = ComponentGetValue2(component, "radius")
							-- if (old_radius < 100) then
								-- ComponentSetValue2(component, "radius", old_radius + 1)
								-- ComponentSetValue2(component, "min_radius", old_radius)
							-- else
								-- ComponentSetValue2(component, "radius", 1)
								-- ComponentSetValue2(component, "min_radius", 0)
							-- end
						-- end
					-- end
					
					for i, child_of_child in ipairs(EntityGetAllChildren(torso_id)) do
						if EntityGetName(child_of_child) == "neck" then
							EntitySetComponentsWithTagEnabled(child_of_child, "enabled_in_world", true)
							neck_id = child_of_child
							nx, ny = EntityGetTransform(neck_id)
							
							local angle = get_direction(nx, ny, x, y)
							local distance = distance(nx, ny, x, y)

							local sprite_width = 34
							local scale = math.max(0, distance) / sprite_width

							EntitySetTransform(neck_id, tx, ty, angle, scale, 1)
							table.insert(neck_entities, neck_id)
						end
					end
				end
			end

			if (times % 200 == 0) and (cooldown <= 0) then
				ComponentSetValue2(var_comp_cooldown, "value_int", 1400)
				for i = 1, 10 do
					local bud_entity = EntityLoad("data/entities/animals/hellion_flower.xml", x, y)
					EntityAddChild(entity_id, bud_entity)
				end
				local worm_component = EntityGetFirstComponentIncludingDisabled( entity_id, "WormComponent")
				ComponentSetValue2( worm_component, "is_water_worm", false )
				ComponentSetValue2( worm_component, "mSpeed", 0.5 )
				ComponentSetValue2( worm_component, "speed", 0.5 )
				ComponentSetValue2( worm_component, "speed_hunt", 0.5 )
				ComponentSetValue2( worm_component, "gravity", 0 )
				ComponentSetValue2( worm_component, "tail_gravity", 10 )
				ComponentSetValue2( worm_component, "max_speed", 25 )
				ComponentSetValue2( worm_component, "mTargetVec", 0, 0 )
				ComponentSetValue2( worm_component, "ground_check_offset", 400 )
				ComponentSetValue2( worm_component, "target_kill_radius", 0 )
			elseif cooldown > 0 then
				ComponentSetValue2(var_comp_cooldown, "value_int", cooldown - 1)
				
				if cooldown == 1380 then
					for i = 1, #neck_entities do
						local sprite_component2 = EntityGetFirstComponentIncludingDisabled(neck_entities[i], "SpriteComponent")
						EntitySetComponentIsEnabled( neck_entities[i], sprite_component2, true )
					end
				end
				
				if cooldown > 500 then
					orbit_radius = 1400 - math.max(cooldown, 1200 )
				elseif cooldown < 200 then
					orbit_radius = math.min(cooldown, 200 )
				end
				if cooldown == 150 then
					for i = 1, #bud_entities do
						EntitySetComponentsWithTagEnabled(bud_entities[i], "enabled_in_hand", false)
						local sprite_component2 = EntityGetFirstComponent(bud_entities[i], "SpriteComponent")
						local an = ComponentGetValue2(sprite_component2, "rect_animation")
						if ( an ~= "close" ) and ( an ~= "despawn" ) then
							ComponentRemoveTag( sprite_component2, "character" )
							ComponentSetValue2(sprite_component2, "rect_animation", "close")
							ComponentSetValue2(sprite_component2, "next_rect_animation", "despawn")
						end
					end
				end
				local angle_offset = times * 0.01  -- Slow rotation
				
				local positions = calculate_orbit_positions(x, y, orbit_radius, angle_offset, 5)
				local positions2 = calculate_orbit_positions(x, y, orbit_radius * 0.5, angle_offset * 0.5, 5)
				for i, pos in ipairs(positions) do
				  EntitySetTransform(bud_entities[i], pos.x, pos.y, pos.angle + (math.pi / 2))
				end

				for i, pos in ipairs(positions2) do
				  EntitySetTransform(bud_entities[i + 5], pos.x, pos.y, pos.angle + (math.pi / 2))
				end
				
				if cooldown == 1200 then
					for i = 1, #bud_entities do
						EntitySetComponentsWithTagEnabled(bud_entities[i], "enabled_in_hand", true)
						local sprite_component2 = EntityGetFirstComponent(bud_entities[i], "SpriteComponent")
						local an = ComponentGetValue2(sprite_component2, "rect_animation")
						if ( an ~= "stand" ) and ( an ~= "open" ) then
							ComponentSetValue2(sprite_component2, "rect_animation", "open")
							ComponentSetValue2(sprite_component2, "next_rect_animation", "stand")
						end
					end
				end
				
				if cooldown == 1 then
					for i = 1, #bud_entities do
						EntityKill(bud_entities[i])
					end
				end
			end
		else
			for i, child in ipairs(EntityGetAllChildren(entity_id) or {}) do
				if EntityGetName(child) == "flower_head" then
					EntitySetComponentsWithTagEnabled(child, "enabled_in_world", true)
					
					for i, child_of_child in ipairs(EntityGetAllChildren(child)) do
						if EntityGetName(child_of_child) == "neck" then
							EntitySetComponentsWithTagEnabled(child_of_child, "enabled_in_world", false)
						end
					end
				end
			end	
		end
	end
end

--Material Converter Circle
local components = EntityGetComponent(entity_id, "MagicConvertMaterialComponent")
if components then
	for k, component in pairs(components)do
		local old_radius = ComponentGetValue2(component, "radius")
		if (old_radius < 100) then
			ComponentSetValue2(component, "radius", old_radius + 1)
			ComponentSetValue2(component, "min_radius", old_radius)
		else
			ComponentSetValue2(component, "radius", 1)
			ComponentSetValue2(component, "min_radius", 0)
		end
	end
end

--Avoid Walls
-- local found, nx, ny, apd = GetSurfaceNormal(x , y, 20, 30)

-- if found then
	-- -- nx = nx * (math.abs(apd))
	-- -- ny = ny * (math.abs(apd))
	
	-- local worm_component = EntityGetFirstComponentIncludingDisabled( entity_id, "WormComponent")
	-- if ( worm_component ~= nil ) and ( worm_component > 0 ) then
		-- ComponentSetValue2( worm_component, "mSpeed", 1 )
		-- ComponentSetValue2( worm_component, "max_speed", 1 )
		-- ComponentSetValue2( worm_component, "mTargetVec", 0, 0 )
	-- end
-- else
	-- nx = 0
	-- ny = 0
	
	-- local worm_component = EntityGetFirstComponentIncludingDisabled( entity_id, "WormComponent")
	-- if ( worm_component ~= nil ) and ( worm_component > 0 ) then
		-- ComponentSetValue2( worm_component, "max_speed", 25 )
	-- end
-- end

x, y, rot = EntityGetTransform(entity_id)
if state <= 1 then
	local did_hit_vert, hit_x, hit_y = RaytraceSurfaces(x, y + 10, x, y - 10)
	local did_hit_up = RaytraceSurfaces(x, y, x, y - 10)
	local did_hit_down = RaytraceSurfaces(x, y, x, y + 10)
	local did_hit_left = RaytraceSurfaces(x, y, x - 10, y)
	local did_hit_right = RaytraceSurfaces(x, y, x + 10, y)

	local did_hit_up_left = RaytraceSurfaces(x, y, x - 10, y - 10)
	local did_hit_up_right = RaytraceSurfaces(x, y, x + 10, y - 10)

	local did_hit_down_left = RaytraceSurfaces(x, y, x - 10, y + 10)
	local did_hit_down_right = RaytraceSurfaces(x, y, x + 10, y + 10)

	if did_hit_down then
		y = y - 1
	end

	if did_hit_up then
		y = y + 1
	end

	if did_hit_left then
		x = x + 1
	end

	if did_hit_right then
		x = x - 1
	end

	if did_hit_up_left then
		x = x + 1
		y = y + 1
	end

	if did_hit_up_right then
		x = x - 1
		y = y + 1
	end

	if did_hit_down_left then
		x = x + 1
		y = y - 1
	end

	if did_hit_down_right then
		x = x - 1
		y = y - 1
		
	end

	EntitySetTransform( entity_id, x, y)
	EntitySetTransform( face_entity, x, y)

	function is_in_wall(x, y)
		return RaytraceSurfaces(x, y, x, y)
	end

	function raytrace_directions(x, y)
		local results = {}
		results.did_hit_vert, results.hit_x, results.hit_y = RaytraceSurfaces(x, y + 10, x, y - 10)
		results.did_hit_up = RaytraceSurfaces(x, y, x, y - 10)
		results.did_hit_down = RaytraceSurfaces(x, y, x, y + 10)
		results.did_hit_left = RaytraceSurfaces(x, y, x - 10, y)
		results.did_hit_right = RaytraceSurfaces(x, y, x + 10, y)

		results.did_hit_up_left = RaytraceSurfaces(x, y, x - 10, y - 10)
		results.did_hit_up_right = RaytraceSurfaces(x, y, x + 10, y - 10)
		results.did_hit_down_left = RaytraceSurfaces(x, y, x - 10, y + 10)
		results.did_hit_down_right = RaytraceSurfaces(x, y, x + 10, y + 10)

		return results
	end

	function adjust_position(x, y, hit_results)
		if hit_results.did_hit_down then
			y = y - 1
		end
		if hit_results.did_hit_up then
			y = y + 1
		end
		if hit_results.did_hit_left then
			x = x + 1
		end
		if hit_results.did_hit_right then
			x = x - 1
		end

		if hit_results.did_hit_up_left then
			x = x + 1
			y = y + 1
		end

		if hit_results.did_hit_up_right then
			x = x - 1
			y = y + 1
		end

		if hit_results.did_hit_down_left then
			x = x + 1
			y = y - 1
		end

		if hit_results.did_hit_down_right then
			x = x - 1
			y = y - 1
		end

		return x, y
	end

	function find_nearest_empty_space(x, y)
		local queue = {{x, y}}
		local visited = {}
		local directions = {
			{1, 0}, {-1, 0}, {0, 1}, {0, -1}, -- cardinal directions (right, left, down, up)
			{1, 1}, {1, -1}, {-1, 1}, {-1, -1} -- diagonal directions
		}
		local raycasts = 8
		local dir = (math.pi * 2.0) / raycasts
		local length = 20

		visited[y] = {}
		visited[y][x] = true

		while #queue > 0 do
			local current = table.remove(queue, 1)
			local cx, cy = current[1], current[2]

			for _, d in ipairs(directions) do
				local nx, ny = cx + d[1], cy + d[2]

				if not (visited[ny] and visited[ny][nx]) then
					visited[ny] = visited[ny] or {}
					visited[ny][nx] = true

					local wall = false
					for j = 0, raycasts - 1 do
						local ex = nx + math.cos(j * dir) * length
						local ey = ny - math.sin(j * dir) * length
						if RaytraceSurfaces(nx, ny, ex, ey) then
							wall = true
							break
						end
					end

					if not wall then
						return nx, ny
					else
						table.insert(queue, {nx, ny})
					end
				end
			end
		end

		-- Fallback to a random position if no empty space found (which should be unlikely)
		return x + Random(-50, 50), y + Random(-50, 50)
	end

	function push_towards(x, y, target_x, target_y, push_amount)
		local push_vector_x = target_x - x
		local push_vector_y = target_y - y

		local length = math.sqrt(push_vector_x^2 + push_vector_y^2)
		if length > 0 then
			push_vector_x = (push_vector_x / length) * push_amount
			push_vector_y = (push_vector_y / length) * push_amount
		end

		return x + push_vector_x, y + push_vector_y
	end

	x, y = EntityGetTransform(entity_id)

	-- local found_normal, normal_x, normal_y, approximate_distance_from_surface = GetSurfaceNormal(x, y, 20, 16)

	local bloppi = false
	if is_in_wall(x, y) then
		local hit_results = raytrace_directions(x, y)
		x, y = adjust_position(x, y, hit_results)
		
		if is_in_wall(x, y) then
			local dx, dy = find_nearest_empty_space(x, y)
			local push_amount = 1 -- Adjust this value to control push speed
			x2, y2 = push_towards(x, y, dx, dy, push_amount)
		end
		
		-- local worm_component = EntityGetFirstComponentIncludingDisabled( entity_id, "WormComponent")
		if ( worm_component ~= nil ) and ( worm_component > 0 ) then
			ComponentSetValue2( worm_component, "mSpeed", 0.5 )
			ComponentSetValue2( worm_component, "max_speed", 0.5 )
			ComponentSetValue2( worm_component, "mTargetVec", 0, 0 )
			bloppi = true
		end

		EntitySetTransform(entity_id, x2, y2)
		EntitySetTransform(face_entity, x2, y2)
	-- elseif found_normal then

		-- local current_speed = ComponentGetValue2(worm_component, "mSpeed")
		-- -- ComponentSetValue2( worm_component, "mSpeed", current_speed * 0.3 )
			-- ComponentSetValue2( worm_component, "mSpeed", 0 )
			-- ComponentSetValue2( worm_component, "max_speed", 0 )
			-- ComponentSetValue2( worm_component, "mTargetVec", 0, 0 )
	end

	if bloppi == false then
		local worm_component = EntityGetFirstComponentIncludingDisabled( entity_id, "WormComponent")
		if ( worm_component ~= nil ) and ( worm_component > 0 ) then
			ComponentSetValue2( worm_component, "max_speed", 25 )
		end
	end
end
--Make enemies weak to new lava
	
if times % 120 == 0 then


	local entities = EntityGetInRadiusWithTag(x, y, 1000, "mortal")

	for i, entity in ipairs(entities) do
		if EntityGetName(entity) ~= "DEBUG_NAME:player" then
			local material_damages = get_material_damage(entity)

			if material_damages and material_damages.acid and material_damages.acid > 0 then

				local entity_damage_model_component = EntityGetFirstComponent(entity, "DamageModelComponent")
				
				if entity_damage_model_component then
				
					EntitySetDamageFromMaterial( entity, "new_enemies_custom_lava", material_damages.acid * 0.15 )
					
				end
				
			end
		
		end

	end
end

--heat
-- local x, y, rot = EntityGetTransform(entity_id)
-- GameSetPostFxParameter("heatmon_pos", x, y, 400, 2)

-- local mask_id = GlobalsGetValue("new_enemies_mod_heat_shader_texture_id", 1)
-- x, y = EntityGetTransform(entity_id)

-- GameCreateSpriteForXFrames( "mods/new_enemies/files/shader_mask/shader_mask.png", x, y, true, 0, 0, 1, true )
-- local mask_id, w, h = ModImageIdFromFilename( "mods/new_enemies/files/shader_mask/shader_mask.png" )
-- local w = 192
-- local h = 108
-- -- Shrink circling path radius
-- x = x * w/4
-- y = y * h/4

-- x = x * h/w -- Aspect correction

-- -- Shift to center of screen
-- x = x + w/2
-- y = y + h/2

-- local r = 40

-- -- Write white pixels to the mask texture in a circle around the effect,
-- if mask_id ~= 0 then
	-- GamePrint(mask_id)
	-- for i = -r, r do
		-- for j = -r, r do
			-- local dist = math.sqrt( i^2 + j^2 )
			-- if dist <= r then
				-- ModImageSetPixel( 1, x + i, y + j, 255 - dist/r * 255 )
			-- end
		-- end
	-- end
-- end
