dofile_once( "data/scripts/invisibility_check.lua" )

local this_component = GetUpdatedComponentID()
local entity_id = GetUpdatedEntityID()
local x, y = EntityGetTransform(entity_id)

local has_player_tag = EntityHasTag(entity_id, "polymorphed_player")
if has_player_tag == true then
	SetValueNumber("time_dashing", 0)
end

local block_script = false

function get_distance(dx, dy, dx2, dy2)
	return math.sqrt(((dx-dx2)^2) + ((dy-dy2)^2))
end

local players = {}
local players_closer = {}
local player = EntityGetWithName( "DEBUG_NAME:player" )
if ( player ~= nil ) and ( player > 0 ) then
	local px, py = EntityGetTransform(player)
	local dist = get_distance(x, y, px, py)
	if dist <= 300 then
		players = {player}
	end
	if dist <= 200 then
		players_closer = {player}
	end
	if dist > 400 then
		block_script = true
	end
else
	local polymorphed_player = EntityGetClosestWithTag(x, y, "polymorphed_player")
	if ( polymorphed_player ~= nil ) and ( polymorphed_player > 0 ) then
		local px, py = EntityGetTransform(polymorphed_player)
		local dist = get_distance(x, y, px, py)
		if dist <= 300 then
			players = {polymorphed_player}
		end
		if dist <= 200 then
			players_closer = {polymorphed_player}
		end
		if dist > 400 then
			block_script = true
		end
	end
end

if block_script == false then
	local components = EntityGetComponent(entity_id, "VariableStorageComponent")

	local pos_comp, vel_comp, momentum_comp, init_comp

	for _, comp_id in pairs(components) do
		local var_name = ComponentGetValue(comp_id, "name")
		if var_name == "position" then
			pos_comp = comp_id
		elseif var_name == "velocity" then
			vel_comp = comp_id
		elseif var_name == "momentum" then
			momentum_comp = comp_id
		elseif var_name == "init" then
			init_comp = comp_id
		elseif var_name == "delay" then
			delay_comp = comp_id
		elseif var_name == "origin" then
			origin_comp = comp_id
		end
	end

	local delay = ComponentGetValue2(delay_comp, "value_int")

	local cx, cy = GameGetCameraPos()
	if (x > cx + 256) or
		(x < cx - 256) or
		(y < cy - 150) or
		(y > cy + 150) then
		ComponentSetValue2(delay_comp, "value_int", 80)
	else
		ComponentSetValue2(delay_comp, "value_int", 1)
	end

	-- GameSetFogOfWar( x, y, 249 )

	SetRandomSeed(GameGetFrameNum(), x + y + entity_id)

	if ComponentGetValue2(init_comp, "value_bool") == false then
		ComponentSetValue2(init_comp, "value_bool", true)
		ComponentSetValue2(pos_comp, "value_string", tostring(x) .. "," .. tostring(y))
		ComponentSetValue2(origin_comp, "value_int", x)
		ComponentSetValue2(origin_comp, "value_float", y)
	end

	local origin_x = ComponentGetValue2(origin_comp, "value_int")
	local origin_y = ComponentGetValue2(origin_comp, "value_float")

	function split_string(input_str, delimiter)
		local result = {}
		for match in (input_str .. delimiter):gmatch("(.-)" .. delimiter) do
			table.insert(result, match)
		end
		return result
	end

	local function is_area_open(test_x, test_y, radius)
		local step_size = 1
		for dx = -radius, radius, step_size do
			for dy = -radius, radius, step_size do
				if RaytracePlatforms(test_x + dx, test_y + dy, test_x + dx, test_y + dy) then
					return false
				end
			end
		end
		return true
	end

	function get_darkest_position(start_x, start_y, grid_size, step_size)
		local darkest_position = { x = start_x, y = start_y }
		local max_score = -math.huge

		for dx = -grid_size / 2, grid_size / 2, step_size do
			for dy = -grid_size / 2, grid_size / 2, step_size do
				local test_x, test_y = start_x + dx, start_y + dy
				local fog_of_war = GameGetFogOfWar(test_x, test_y)
				local biome_name = BiomeMapGetName(test_x, test_y)
				if fog_of_war >= 0 and (biome_name == "Crypt" or biome_name == BiomeMapGetName(origin_x, origin_y) or biome_name == "Holy Temple" or biome_name == "$biome_tower" or biome_name == "$biome_wandcave" or biome_name == "$biome_pyramid" or biome_name == "$biome_crypt") then
					if is_area_open(test_x, test_y, 15) then
						local distance = math.sqrt((test_x - start_x) ^ 2 + (test_y - start_y) ^ 2)
						local score = fog_of_war - distance * 0.1
						if score > max_score then
							max_score = score
							darkest_position = { x = test_x, y = test_y }
						end
					end
				end
			end
		end

		return darkest_position
	end

	function move_away_from_light()
		local grid_size = 300
		local step_size = 20

		local darkest_position = get_darkest_position(x, y, grid_size, step_size)

		local pos = { x = x, y = y }
		local vel = { x = 0, y = 0 }
		local momentum = { x = 0, y = 0 }

		local pos_split = { string.match(ComponentGetValue2(pos_comp, "value_string"), "([^,]+),([^,]+)") }
		local vel_split = { string.match(ComponentGetValue2(vel_comp, "value_string"), "([^,]+),([^,]+)") }
		local momentum_split = { string.match(ComponentGetValue2(momentum_comp, "value_string"), "([^,]+),([^,]+)") }

		if #pos_split == 2 then pos.x, pos.y = tonumber(pos_split[1]), tonumber(pos_split[2]) end
		if #vel_split == 2 then vel.x, vel.y = tonumber(vel_split[1]), tonumber(vel_split[2]) end
		if #momentum_split == 2 then momentum.x, momentum.y = tonumber(momentum_split[1]), tonumber(momentum_split[2]) end

		local dir_x, dir_y = darkest_position.x - pos.x, darkest_position.y - pos.y
		local length = math.sqrt(dir_x ^ 2 + dir_y ^ 2)
		if length > 0 then
			dir_x, dir_y = dir_x / length, dir_y / length
		end

		local speed = 0.2
		vel.x, vel.y = vel.x * 0.9 + dir_x * speed * Random(0.5, 1), vel.y * 0.9 + dir_y * speed * Random(0.5, 1)

		local max_speed = 5
		local vel_length = math.sqrt(vel.x ^ 2 + vel.y ^ 2)
		if vel_length > max_speed then
			vel.x, vel.y = vel.x / vel_length * max_speed, vel.y / vel_length * max_speed
		end

		pos.x, pos.y = pos.x + vel.x + momentum.x * Random(0.5, 1), pos.y + vel.y + momentum.y * Random(0.5, 1)

		ComponentSetValue2(pos_comp, "value_string", string.format("%f,%f", pos.x, pos.y))
		ComponentSetValue2(vel_comp, "value_string", string.format("%f,%f", vel.x, vel.y))
		ComponentSetValue2(momentum_comp, "value_string", string.format("%f,%f", momentum.x, momentum.y))

		EntitySetTransform(entity_id, pos.x, pos.y)
	end

	function dash_towards_player(player_x, player_y)
		local pos_str = ComponentGetValue2(pos_comp, "value_string")
		local vel_str = ComponentGetValue2(vel_comp, "value_string")
		local momentum_str = ComponentGetValue2(momentum_comp, "value_string")

		local pos_split = { string.match(pos_str, "([^,]+),([^,]+)") }
		local vel_split = { string.match(vel_str, "([^,]+),([^,]+)") }
		local momentum_split = { string.match(momentum_str, "([^,]+),([^,]+)") }

		local pos = { x = tonumber(pos_split[1]), y = tonumber(pos_split[2]) }
		local vel = { x = tonumber(vel_split[1]), y = tonumber(vel_split[2]) }
		local momentum = { x = tonumber(momentum_split[1]), y = tonumber(momentum_split[2]) }

		local dir_x, dir_y = player_x - pos.x, player_y - pos.y
		local length = math.sqrt(dir_x ^ 2 + dir_y ^ 2)
		if length > 0 then
			dir_x, dir_y = dir_x / length, dir_y / length
		end

		local dash_speed = 0.5
		vel.x, vel.y = dir_x * dash_speed, dir_y * dash_speed

		local max_speed = 5
		local vel_length = math.sqrt(vel.x ^ 2 + vel.y ^ 2)
		if vel_length > max_speed then
			vel.x, vel.y = vel.x / vel_length * max_speed, vel.y / vel_length * max_speed
		end

		pos.x, pos.y = pos.x + vel.x + momentum.x * Random(0.9, 1.5), pos.y + vel.y + momentum.y * Random(0.9, 1.5)

		ComponentSetValue2(pos_comp, "value_string", string.format("%f,%f", pos.x, pos.y))
		ComponentSetValue2(vel_comp, "value_string", string.format("%f,%f", vel.x, vel.y))
		ComponentSetValue2(momentum_comp, "value_string", string.format("%f,%f", momentum.x, momentum.y))

		EntitySetTransform(entity_id, pos.x, pos.y)
	end

	local did_hit, hit_x, hit_y = RaytracePlatforms(x - 5, y, x + 5, y)
	local did_hit2, hit_x, hit_y = RaytracePlatforms(x - 5, y - 5, x + 5, y + 5)
	local did_hit3, hit_x, hit_y = RaytracePlatforms(x, y - 5, x, y + 5)
	local did_hit4, hit_x, hit_y = RaytracePlatforms(x - 5, y + 5, x + 5, y - 5)
	local fog_of_war = GameGetFogOfWar(x, y)
	if ( fog_of_war <= 250 ) or ( did_hit ) or ( did_hit2 ) or ( did_hit3 ) or ( did_hit4 ) then
		move_away_from_light()
	end

	local times_executed = ComponentGetValue2(GetUpdatedComponentID(), "mTimesExecuted")
	local time_exposed = GetValueNumber("time_exposed", 0)
	if fog_of_war <= 250 then
		
		SetValueNumber("time_exposed", time_exposed + 1)
	else
		SetValueNumber("time_exposed", 0)
	end
		
	if times_executed % delay == 0 then


		local time_dashing = GetValueNumber("time_dashing", 0)

		if ( times_executed % 600 == 0 ) and ( #players_closer > 0 ) and ( EntityGetFirstComponent(players[1], "GenomeDataComponent") ~= nil ) and ( EntityGetFirstComponent(players[1], "GenomeDataComponent") > 0 ) and ( EntityGetHerdRelation(entity_id, players[1]) < 40 ) and ( is_invisible(players[1]) == false ) then
			SetValueNumber("time_dashing", 160)
		end

		if time_dashing > 0 then
			
			if #players > 0 then
				local px, py = EntityGetTransform(players[1])
				dash_towards_player(px, py)
				EntitySetComponentsWithTagEnabled(entity_id, "sound_prebattle_tinkering", true)
				EntitySetComponentsWithTagEnabled(entity_id, "enabled_in_world", true)
				EntitySetComponentsWithTagEnabled(entity_id, "enabled_in_hand", false)
				EntitySetComponentsWithTagEnabled(entity_id, "pitcheck_b", false)
				GameSetFogOfWar( x, y, 255 )
				GamePlaySound("data/audio/Desktop/animals.bank", "animals/maggot/death", x, y)
			end
			SetValueNumber("time_dashing", time_dashing - 1)
			
			SetValueNumber("time_exposed", 0)
		else
			EntitySetComponentsWithTagEnabled(entity_id, "sound_prebattle_tinkering", false)
			if time_exposed > 90 then

				if times_executed % 10 == 0 then
					GamePlaySound("mods/new_enemies/data/sound/more_sounds/my_mod_audio.bank", "snd_mod/hiss", x, y)
					if #players > 0 then
						local comp = EntityGetFirstComponentIncludingDisabled( entity_id, "DamageModelComponent" )
						if ( comp ~= nil ) then
							local max_hp = ComponentGetValue2( comp, "max_hp" )
							local hp = ComponentGetValue2( comp, "hp" )
							EntityInflictDamage( entity_id, max_hp / 40, "DAMAGE_PROJECTILE", "", "NORMAL", 0, 0, players[1], x, y )
						end
					end

				end
				EntitySetComponentsWithTagEnabled(entity_id, "enabled_in_world", true)
				EntitySetComponentsWithTagEnabled(entity_id, "enabled_in_hand", false)
				EntitySetComponentsWithTagEnabled(entity_id, "pitcheck_b", true)
			else
				EntitySetComponentsWithTagEnabled(entity_id, "enabled_in_world", false)
				EntitySetComponentsWithTagEnabled(entity_id, "enabled_in_hand", true)
				EntitySetComponentsWithTagEnabled(entity_id, "pitcheck_b", false)
			end
		end
	end

	--SAUSAGE LEGS:
	--Moved into here for optimization!
	function draw_sosig_limb_function(entity)
		function approx_rolling_average(avg, target, n)
			avg = avg - avg / n;
			avg = avg + target / n;
			return avg
		end
		local smoothness_factor = 20

		local function distance(x1, y1, x2, y2)
		  local dx = x2 - x1
		  local dy = y2 - y1
		  return math.sqrt(dx * dx + dy * dy)
		end
		-- Function to subdivide a 2D segmented line into parts of equal length
		function subdivide_line(segmented_line, equal_length)
		  if #segmented_line < 2 or equal_length <= 0 then
			  return nil
		  end

		  local total_length = 0
		  for i = 1, #segmented_line - 1 do
			  local x1, y1 = segmented_line[i].x, segmented_line[i].y
			  local x2, y2 = segmented_line[i + 1].x, segmented_line[i + 1].y
			  total_length = total_length + distance(x1, y1, x2, y2)
		  end

		  local num_subdivisions = math.floor(total_length / equal_length)
		  if num_subdivisions == 0 then
			  return nil
		  end

		  local segment_length = total_length / num_subdivisions

		  local result = {}
		  table.insert(result, segmented_line[1]) -- Include the starting point

		  local current_length = 0
		  for i = 1, #segmented_line - 1 do
			  local x1, y1 = segmented_line[i].x, segmented_line[i].y
			  local x2, y2 = segmented_line[i + 1].x, segmented_line[i + 1].y
			  local segment_dist = distance(x1, y1, x2, y2)

			  while current_length + segment_dist >= segment_length do
				  local t = (segment_length - current_length) / segment_dist
				  local new_x = x1 + t * (x2 - x1)
				  local new_y = y1 + t * (y2 - y1)
				  table.insert(result, { x = new_x, y = new_y })
				  segment_dist = segment_dist - (segment_length - current_length)
				  x1, y1 = new_x, new_y
				  current_length = 0
			  end

			  current_length = current_length + segment_dist
		  end

		  -- Include the ending point
		  table.insert(result, segmented_line[#segmented_line])

		  return result
		end

		function map(value, old_min, old_max, new_min, new_max)
			return (value - old_min) * (new_max - new_min) / (old_max - old_min) + new_min;
		end

		-- Function to calculate the arc points with adjustable amplitude
		function calculate_arc(start_x, start_y, end_x, end_y, num_points, concave, amplitude)
			local points = {}
			local mid_x = (start_x + end_x) / 2
			local mid_y = (start_y + end_y) / 2

			local control_x = mid_x
			local control_y = mid_y + (concave and -amplitude or amplitude)  -- Use amplitude to control the concavity

			for t = 0, 1, 1 / (num_points - 1) do
				local u = 1 - t
				local tt = t * t
				local uu = u * u

				local x = uu * start_x + 2 * u * t * control_x + tt * end_x
				local y = uu * start_y + 2 * u * t * control_y + tt * end_y
				table.insert(points, { x = x, y = y })
			end

			return points
		end

		-- Function to draw limb using sprite components
		function draw_limb(entity_id, start_x, start_y, end_x, end_y, scale_x, amplitude, cancel)
			local num_points = 10  -- Adjust the number of points for smoothness
			local points = calculate_arc(start_x, start_y, end_x, end_y, num_points, scale_x > 0, amplitude)
			local half_point = #points / 2
			local sprite_size = 9
			local max_length = 100

			-- Calculate total length of the arc
			local path_length = 0
			for i = 1, #points - 1 do
				path_length = path_length + distance(points[i].x, points[i].y, points[i + 1].x, points[i + 1].y)
			end

			-- Adjust scaling
			local dist_p = math.min(1, path_length / max_length)
			dist_p = map(dist_p, 0, 1, 3.5, 2)
			local new_path = subdivide_line(points, dist_p * cancel) or {}

			-- Create or remove SpriteComponents until it matches the amount of points we have on the path
			local sprite_comps = EntityGetComponentIncludingDisabled(entity_id, "SpriteComponent") or {}
			-- Create new comps until we have enough
			for i = 1, #new_path - #sprite_comps do
				EntityAddComponent2(entity_id, "SpriteComponent", {
					image_file = "data/enemies_gfx/manticore_segment.png",
					update_transform_rotation = false,
					has_special_scale = true,
					z_index = 8.1
				})
			end
			-- Remove comps that are too many
			for i = 1, #sprite_comps - #new_path do
				EntityRemoveComponent(entity_id, sprite_comps[i])
			end
			sprite_comps = EntityGetComponentIncludingDisabled(entity_id, "SpriteComponent") or {}

			-- Set SpriteComponent offsets
			for i, v in ipairs(new_path) do
				local h = (i - half_point) / #new_path
				h = h * h
				local scale = 1 - math.min(max_length, path_length) / max_length
				scale = scale + h * 2
				scale = math.min(1, math.max(0.8, scale))

				ComponentSetValue2(sprite_comps[i], "offset_x", -(v.x - start_x) * (1 / scale) + (sprite_size / 2))
				ComponentSetValue2(sprite_comps[i], "offset_y", -(v.y - start_y) * (1 / scale) + (sprite_size / 2))
				ComponentSetValue2(sprite_comps[i], "special_scale_x", scale)
				ComponentSetValue2(sprite_comps[i], "special_scale_y", scale)
			end
		end
		
		local components = EntityGetComponent(entity, "VariableStorageComponent")

		local pos_comp, vel_comp, momentum_comp, init_comp

		for _, comp_id in pairs(components) do
			local var_name = ComponentGetValue(comp_id, "name")
			if var_name == "targ_x" then
				targ_x_comp = comp_id
			elseif var_name == "targ_y" then
				targ_y_comp = comp_id
			end
		end

		-- Integrate the draw_limb function with the IK limb script
		local ex, ey, r, sx = EntityGetTransform(entity)

		local target = {ComponentGetValue2(targ_x_comp, "value_float"), ComponentGetValue2(targ_y_comp, "value_float")}

		-- Use the draw_limb function for each limb segment with reduced amplitude
		local amplitude = 25  -- Adjust this value to control the arc height
		if times_executed > 100 then
			if ( #players > 0 ) then
				draw_limb(entity, ex, ey, target[1], target[2] + 20, -1, amplitude, 1)
			else
				draw_limb(entity, ex, ey, ex, ey, -1, amplitude, 500)
			end
		end


		local ik_limb_comp = EntityGetFirstComponentIncludingDisabled(entity, "IKLimbWalkerComponent")
		local new_x, new_y = ComponentGetValue2(ik_limb_comp, "mTarget")

		if new_y < ey then
			new_y = ey + (ey - new_y)
		end

		local average_x = approx_rolling_average(target[1], new_x, smoothness_factor)
		local average_y = approx_rolling_average(target[2], new_y, smoothness_factor)
		
		ComponentSetValue2(targ_x_comp, "value_float", average_x)
		ComponentSetValue2(targ_y_comp, "value_float", average_y)
	end

	-- if times_executed % delay == 0 then
		-- draw_sosig_limb_function(entity_id)
	-- end

	--MEDUSA HAIR:
	-- Define the radius and the animation speed
	local radius_x = 4  -- Width of the oval
	local radius_y = 2   -- Height of the oval
	local animation_speed = 0.02

	-- Retrieve all child entities
	if times_executed % delay == 0 then

		for i, child_id in ipairs(EntityGetAllChildren(entity_id) or {}) do
		
			if EntityGetName(child_id) == "medusa_snake" then
			
				local x, y = EntityGetTransform(child_id)
				
				-- Use the entity ID to generate a unique phase offset for each snake
				local phase_offset = child_id * 1

				-- Calculate the new position using sine and cosine functions
				local angle = (times_executed * animation_speed) + phase_offset
				local offset_x = math.cos(angle) * radius_x
				local offset_y = math.sin(angle) * radius_y

				local verlet_component = EntityGetFirstComponentIncludingDisabled(child_id, "VerletPhysicsComponent")
				if verlet_component then
					ComponentSetValue2(verlet_component, "animation_target_offset", offset_x, offset_y - 6)
				end
				
			end
			
			--Sosig legs handling
		
			if EntityGetName(child_id) == "manticore_body1" then
				for i, child_id2 in ipairs(EntityGetAllChildren(child_id) or {}) do
					if EntityGetName(child_id2) == "sausage_limb" then
						draw_sosig_limb_function(child_id2)
					end
					if EntityGetName(child_id2) == "manticore_body2" then
						for i, child_id3 in ipairs(EntityGetAllChildren(child_id2) or {}) do
							if EntityGetName(child_id3) == "sausage_limb" then
								draw_sosig_limb_function(child_id3)
							end
						end
					end
				end
			end
		end
	end
end

local controls_override = false
local has_player_tag = EntityHasTag(entity_id, "polymorphed_player")
if has_player_tag == true then
	local comps = EntityGetAllComponents(entity_id)
	local concomp = nil
	for i,v in ipairs(comps) do
		if ComponentGetTypeName(v) == "ControlsComponent" then concomp = v; break end
	end
	local input = {
		ComponentGetValue(concomp, "mButtonDownRight"), --right
		ComponentGetValue(concomp, "mButtonDownUp"), --up
		ComponentGetValue(concomp, "mButtonDownLeft"), --left
		ComponentGetValue(concomp, "mButtonDownDown"), --down
	}
	
	local tx = x
	local ty = y
	
	if input[1] == "1" then tx = tx + 100 end
	if input[2] == "1" then ty = ty - 100 end
	if input[3] == "1" then tx = tx - 100 end
	if input[4] == "1" then ty = ty + 100 end
	dash_towards_player(tx, ty)
else
	if block_script == true then
		ComponentSetValue(this_component, "execute_every_n_frame", 150)
	else
		ComponentSetValue(this_component, "execute_every_n_frame", 1)
	end
end