function collision_trigger(colliding_entity)
	function get_fungal_shifted_material_name(material_name, multiple_materials_boolean)
		-- Get the shifts if not provided
		local world_state_entity_id = GameGetWorldStateEntity()
		local world_state_component = EntityGetFirstComponentIncludingDisabled(world_state_entity_id, "WorldStateComponent")
		local shifts = ComponentGetValue2(world_state_component, "changed_materials")

		-- Function to find shifts for a single material
		function find_shifts(target)
			local results = {}
			for i = #shifts - 1, 1, -2 do
				local from = shifts[i]
				local to = shifts[i + 1]

				if target == from or target == to then
					table.insert(results, {from, to})
				end
			end
			return results
		end

		if multiple_materials_boolean and type(material_name) == "table" then
			-- Handle multiple materials
			local all_results = {}
			for _, name in ipairs(material_name) do
				local results = find_shifts(name)
				if #results > 0 then
					all_results[name] = results
				end
			end
			return all_results
		else
			-- Handle single material
			local results = find_shifts(material_name)
			if #results > 0 then
				return results[1][1], results[1][2]
			end
		end
	end

	local from, to = get_fungal_shifted_material_name("magic_liquid_random_polymorph")
	local has_shifted = false
	if from == "magic_liquid_random_polymorph" then
		ConvertMaterialEverywhere( "magic_liquid_random_polymorph_temp", to )
		has_shifted = true
	end

	from, to = get_fungal_shifted_material_name("magic_liquid_polymorph")
			
	if from == "magic_liquid_polymorph" then
		ConvertMaterialEverywhere( "magic_liquid__polymorph_temp", to )
		has_shifted = true
	end

	from, to = get_fungal_shifted_material_name("radioactive_liquid")
			
	if from == "radioactive_liquid" then
		ConvertMaterialEverywhere( "radioactive_liquid_blue", to )
		has_shifted = true
	end
	
	if has_shifted == true then
		local entity_id = GetUpdatedEntityID()
		EntityRemoveComponent( entity_id, GetUpdatedComponentID() )
	end
end