dofile_once( "data/scripts/invisibility_check.lua" )

local times_executed = ComponentGetValue2(GetUpdatedComponentID(), "mTimesExecuted")

function is_valid_entity( entity_id )
	return entity_id ~= nil and entity_id ~= 0
end

function get_variable_storage_component( entity_id, name )
	if not is_valid_entity( entity_id ) then  
		return nil
	end

	local components = EntityGetComponent( entity_id, "VariableStorageComponent" )
	if( components ~= nil ) then
		for _,comp_id in pairs(components) do 
			local var_name = ComponentGetValue( comp_id, "name" )
			if( var_name == name) then
				return comp_id
			end
		end
	end

	return nil
end
		
local entity_id = GetUpdatedEntityID()

local worm_ai_component = EntityGetFirstComponentIncludingDisabled( entity_id, "WormAIComponent")

local worm_component = EntityGetFirstComponentIncludingDisabled( entity_id, "WormComponent")
local velx, vely = ComponentGetValue2(worm_component, "mTargetVec")

local velocity_magnitute = math.sqrt(velx^2 + vely^2)

-- if worm_ai_component then
	-- local players = EntityGetWithTag("player_unit")
	-- ComponentSetValue2( worm_ai_component, "mTargetEntityId", players[1] )
-- end

local x, y = EntityGetTransform(entity_id)

SetRandomSeed( GameGetFrameNum(), x + y + entity_id )

local segment_names = {
  "sand_worm_segment",
}
local segments = {}

-- table.insert(segments, entity_id)

local is_performing_flash_delay = false

local target_id = 0

for _, child in ipairs(EntityGetAllChildren(entity_id) or {}) do
  local child_name = EntityGetName(child)
  if child_name == "target" then
	target_id = child
  end
  for _, segment_name in ipairs(segment_names) do
    if child_name == segment_name then
      table.insert(segments, child)
		local var_comp = EntityGetFirstComponent(child, "VariableStorageComponent")
		if ComponentGetValue2(var_comp, "value_bool") == false then
			ComponentSetValue2(var_comp, "value_bool", true)
			
			local var_comp2 = get_variable_storage_component( child, "origin" )
			ComponentSetValue2(var_comp2, "value_int", x)
			ComponentSetValue2(var_comp2, "value_string", y)
		end
		
		local var_comp5 = get_variable_storage_component( child, "damage_flash_delay" )
		local flash_delay = ComponentGetValue2(var_comp5, "value_int")
		if flash_delay > 0 then
			is_performing_flash_delay = true
			ComponentSetValue2(var_comp5, "value_int", flash_delay - 1)
		end
      break
    end
  end
end

function get_distance(x1, y1, x2, y2)
  local result = math.sqrt((x2 - x1) ^ 2 + (y2 - y1) ^ 2)
  return result
end

function get_direction(x1, y1, x2, y2)
  return math.atan2((y2 - y1), (x2 - x1))
end

function parallax(entity1, entity2, factor)
	-- local var_comp = EntityGetFirstComponent(entity_id, "VariableStorageComponent")
	-- if ComponentGetValue2(var_comp, "value_bool") == false then
		-- ComponentSetValue2(var_comp, "value_bool", true)
		
		-- local var_comp2 = get_variable_storage_component( entity_id, "origin" )
		-- ComponentSetValue2(var_comp2, "value_int", x)
		-- ComponentSetValue2(var_comp2, "value_string", y)
		
		-- SetValueNumber("origin_x", x)
		-- SetValueNumber("origin_y", y)
	-- end

	-- local ox = GetValueNumber("origin_x", x)
	-- local oy = GetValueNumber("origin_y", y)
	local var_comp1 = get_variable_storage_component( entity1, "origin" )
	local ox1 = ComponentGetValue2(var_comp1, "value_int")
	local oy1 = tonumber(ComponentGetValue2(var_comp1, "value_string"))

	local cx, cy = GameGetCameraPos()

	local parallaxFactor = factor

	local x2 = (cx - ox1) * parallaxFactor
	local y2 = (cy - oy1) * parallaxFactor

	return x2, y2, ox1, oy1
end

function map(s, a1, a2, b1, b2)
  return b1 + (s-a1)*(b2-b1)/(a2-a1)
end

function constrain_and_resize(entity1, entity2, max_dist, parallax_factor)
	
	local x1, y1 = EntityGetTransform(entity1)
	local x2, y2 = EntityGetTransform(entity2)
	
	if (not x1) or (not x2) then return nil end
	
	local dir = get_direction(x1, y1, x2, y2)

	local parax, paray, ox1, oy1 = parallax(entity1, entity2, map(parallax_factor, 0.3, 1, 0.3, 0))
	local field_of_view = 5
	local scale = field_of_view / (field_of_view + ( parallax_factor / 8 ))

	-- local visual_scale = map(scale, 0, 1, 0.1, 1)

	EntitySetTransform(entity1, ox1 + parax, oy1 + paray, dir, parallax_factor, parallax_factor)
	
	local sprite_component = EntityGetFirstComponentIncludingDisabled( entity1, "SpriteComponent")
	ComponentSetValue2(sprite_component, "z_index", 1.2 - parallax_factor)
	
	-- if ( ( 0.8 - parallax_factor > 0.3 ) and (0.8 - previous_parallax_factor <= 0.3) ) or ( ( 0.8 - parallax_factor <= 0.3 ) and (0.8 - previous_parallax_factor > 0.3) ) then
		-- EntityLoad( "data/entities/particles/hell_overseer_stepping_particles.xml", x1, y1 )
		-- GamePlaySound("data/audio/Desktop/animals.bank", "animals/fireskull/death", x1, y1)
	-- end
	
	if 0.8 - parallax_factor > 0.3 then
		ComponentSetValue2(sprite_component, "visible", false)
		EntitySetComponentsWithTagEnabled(entity1, "enabled_in_hand", true)
	else
		ComponentSetValue2(sprite_component, "visible", true)
		EntitySetComponentsWithTagEnabled(entity1, "enabled_in_hand", false)
	end
	if is_performing_flash_delay == false then
		EntityRefreshSprite( entity1, sprite_component )
	end
	
	if ( 1.2 - parallax_factor ) < 0.6 then
		EntitySetComponentsWithTagEnabled(entity1, "enabled_in_world", true)
	else
		EntitySetComponentsWithTagEnabled(entity1, "enabled_in_world", false)
	end

	-- EntitySetTransform(entity1, target_x, target_y, dir, visual_scale, visual_scale)
end

function update_origin(entity1, entity2, max_distance, origin_var_comp)
	local x1 = ComponentGetValue2(origin_var_comp, "value_int")
	local y1 = tonumber(ComponentGetValue2(origin_var_comp, "value_string"))
	
	local var_comp2 = get_variable_storage_component( entity2, "origin" )
	local x2 = ComponentGetValue2(var_comp2, "value_int")
	local y2 = tonumber(ComponentGetValue2(var_comp2, "value_string"))

	local dist = math.sqrt(((x2-x1)^2) + ((y2-y1)^2))
	if dist > max_distance then
		ComponentSetValue2(origin_var_comp, "value_int", x2 + (((x1-x2)/dist)*max_distance))
		ComponentSetValue2(origin_var_comp, "value_string", y2 + (((y1-y2)/dist)*max_distance))
	end
end

function constrain_and_resize_head(entity1, damage_radius, parallax_factor, index, phase_direction_var_comp, previous_phase_var_comp, origin_var_comp)
	-- local velx, vely = GameGetVelocityCompVelocity( entity_id )
	local previous_phase = ComponentGetValue2(previous_phase_var_comp, "value_float")
	local previous_parallax_factor = previous_phase
	
	local phase_direction = ComponentGetValue2(phase_direction_var_comp, "value_float")
	
	local x1, y1 = EntityGetTransform(entity1)
	
	local dir = get_direction(x, y, x + velx, y + vely)

	local parax, paray, ox1, oy1 = parallax(entity1, _, map(parallax_factor, 0.3, 1, 0.3, 0))
	local field_of_view = 5
	local scale = field_of_view / (field_of_view + ( parallax_factor / 8 ))

	-- local visual_scale = map(scale, 0, 1, 0.1, 1)

	EntitySetTransform(entity1, x + parax, y + paray, dir, -parallax_factor, parallax_factor)
	
	local sprite_component = EntityGetFirstComponentIncludingDisabled( entity1, "SpriteComponent")
	
	ComponentSetValue2(sprite_component, "z_index", 1.2 - parallax_factor)

	if ( ( 0.8 - parallax_factor > 0.3 ) and (0.8 - previous_parallax_factor <= 0.3) ) or ( ( 0.8 - parallax_factor <= 0.3 ) and (0.8 - previous_parallax_factor > 0.3) ) then
		GamePlaySound("data/audio/Desktop/explosion.bank", "explosions/barrel_oil", x1, y1)
	end
	
	if ( 0.8 - parallax_factor > 0.3 ) then
		ComponentSetValue2(sprite_component, "visible", false)
		EntitySetComponentsWithTagEnabled(entity1, "enabled_in_hand", true)
	else
		ComponentSetValue2(sprite_component, "visible", true)
		EntitySetComponentsWithTagEnabled(entity1, "enabled_in_hand", false)
	end

	if ComponentGetValue2(GetUpdatedComponentID(), "mTimesExecuted") % (5 - math.min(4, math.floor(velocity_magnitute / 5))) == 0 then
		-- Extract animation name and number
		local current_animation = ComponentGetValue2(sprite_component, "rect_animation")
		local animation_number = tonumber(string.match(current_animation, "%d+"))
		if animation_number then
			-- Update animation value for looping
			animation_number = animation_number + 1
			if animation_number > 7 then
			  animation_number = 1
			end
			
			if animation_number == 4 and ( ( 1.2 - parallax_factor ) < 0.6 ) then
				local entities = EntityGetInRadiusWithTag( x1, y1, damage_radius, "mortal" )
				for i,entity in ipairs( entities ) do
					if ( entity ~= entity_id ) and ( EntityGetName(entity) ~= "sand_worm_segment" ) and ( entity ~= 0 ) and ( entity ~= nil ) then
						local ex, ey = EntityGetTransform(entity)
						EntityInflictDamage( entity, 1, "DAMAGE_SLICE", "Sand Worm Bite", "BLOOD_EXPLOSION", 0, 0, entity_id, ex, ey, 0)
						EntityInflictDamage( entity, 1, "DAMAGE_MELEE", "Sand Worm Bite", "BLOOD_EXPLOSION", 0, 0, entity_id, ex, ey, 0)
					end
				end
				GamePlaySound("data/audio/Desktop/animals.bank", "animals/giantshooter/attack_bite", x1, y1)
			end
			
			if phase_direction < 0 then
				next_animation_name = "mouth" .. tostring(animation_number)
			else
				next_animation_name = "mouthb" .. tostring(animation_number)
			end
			
			-- if ( 0.8 - parallax_factor > 0.25 ) then
				-- next_animation_name = "mouthb" .. tostring(animation_number)
			-- elseif ( 0.8 - parallax_factor > 0.2 ) then
				-- next_animation_name = "mouthd" .. tostring(animation_number)
			-- elseif ( 0.8 - parallax_factor > 0 ) then
				-- next_animation_name = "mouthe" .. tostring(animation_number)
			-- elseif ( 0.8 - parallax_factor > -0.05 ) then
				-- next_animation_name = "mouthc" .. tostring(animation_number)
			-- elseif ( 0.8 - parallax_factor > -0.1 ) then
				-- next_animation_name = "mouth" .. tostring(animation_number)
			-- end

			ComponentSetValue2(sprite_component, "rect_animation", next_animation_name)
			ComponentSetValue2(sprite_component, "next_rect_animation", next_animation_name)
		end
	end
	if is_performing_flash_delay == false then
		EntityRefreshSprite( entity1, sprite_component )
	end
	if ( 1.2 - parallax_factor ) < 0.6 then
		EntitySetComponentsWithTagEnabled(entity1, "enabled_in_world", true)
		
		--damage
		-- if ComponentGetValue2(GetUpdatedComponentID(), "mTimesExecuted") % 30 == 0 then
		-- end
	else
		EntitySetComponentsWithTagEnabled(entity1, "enabled_in_world", false)
	end
	
	ComponentSetValue2(origin_var_comp, "value_int", x)
	ComponentSetValue2(origin_var_comp, "value_string", y)
end

for i = 1, #segments do

	local components = EntityGetComponent( segments[i], "VariableStorageComponent" )
	if ( components ~= nil ) then
		for _,comp_id in pairs(components) do 
			local var_name = ComponentGetValue( comp_id, "name" )
			if ( var_name == "phase") then
				phase_var_comp = comp_id
			end
			if ( var_name == "phase_direction") then
				phase_direction_var_comp = comp_id
			end
			if ( var_name == "previous_phase") then
				previous_phase_var_comp = comp_id
			end
			if ( var_name == "origin") then
				origin_var_comp = comp_id
			end
		end
	end
	
	local p = ComponentGetValue2(phase_var_comp, "value_float")
	p = p + (1/500) % 1
	ComponentSetValue2(phase_var_comp, "value_float", p)
	
	local value = map(math.sin(-p * math.pi * 2 + (i / #segments) * math.pi), -1, 1, 0.3, 1)
	ComponentSetValue2(phase_direction_var_comp, "value_float", math.cos(-p * math.pi * 2 + (i / #segments) * math.pi))
	ComponentSetValue2(previous_phase_var_comp, "value_float", value)
	
	if i == 1 then
		constrain_and_resize_head(segments[i], 35 * value, value, i , phase_direction_var_comp, previous_phase_var_comp, origin_var_comp)
	else
		update_origin(segments[i], segments[i - 1], 24 * value, origin_var_comp)
		constrain_and_resize(segments[i], segments[i - 1], 24 * value, value, i, origin_var_comp)
		
	end
end

local going_back_to_home_var_comp = get_variable_storage_component( entity_id, "going_back_to_home" )
local is_going_back_to_home = ComponentGetValue2(going_back_to_home_var_comp, "value_int")
-- GamePrint(is_going_back_to_home)
local origin_var_comp = get_variable_storage_component( entity_id, "origin" )
local origin_x = ComponentGetValue2(origin_var_comp, "value_int")
local origin_y = ComponentGetValue2(origin_var_comp, "value_float")
local tx, ty = EntityGetTransform( target_id )
local hydra = EntityGetWithName( "$hydra" )
if ComponentGetValue2(origin_var_comp, "value_int") ~= 0 then
	local distance = get_distance(x, y, origin_x, origin_y)
	local worm_ai_component = EntityGetFirstComponentIncludingDisabled( entity_id, "WormAIComponent")
	-- if distance > 2500 then
	local players = {}
	local player = EntityGetWithName( "DEBUG_NAME:player" )
	if ( player ~= nil ) and ( player > 0 ) then
		players = {player}
	else
		local polymorphed_player = EntityGetClosestWithTag(x, y, "polymorphed_player")
		if ( polymorphed_player ~= nil ) and ( polymorphed_player > 0 ) then
			players = {polymorphed_player}
		end
	end
	if ( worm_ai_component ~= nil ) and ( worm_ai_component > 0 ) then
		if ( hydra ~= nil ) and ( hydra > 0 ) then
			local worm_component = EntityGetFirstComponentIncludingDisabled( entity_id, "WormComponent")
			if ( worm_component ~= nil ) and ( worm_component > 0 ) then
				local hx, hy = EntityGetTransform( hydra )
				local distance_hydra = get_distance(x, y, hx, hy)
				if distance_hydra < 450 then
					-- Calculate direction vector from hydra to worm
					local dx = x - hx
					local dy = y - hy
					
					-- Normalize the direction vector
					local magnitude = math.sqrt(dx * dx + dy * dy)
					if magnitude > 0 then
						dx = dx / magnitude
						dy = dy / magnitude
					end
					
					-- Scale the normalized vector by the desired force
					local force_magnitude = 0.1 -- Adjust this value to set the strength of the push
					local force_x = dx * force_magnitude
					local force_y = dy * force_magnitude
					
					-- Apply the force to the worm's target vector
					local vx, vy = ComponentGetValue2(worm_component, "mTargetVec")
					ComponentSetValue2(worm_component, "mTargetVec", vx + force_x, vy + force_y)
				end
			end
		elseif is_going_back_to_home == 1 then
			local worm_component = EntityGetFirstComponentIncludingDisabled( entity_id, "WormComponent")
			if ( worm_component ~= nil ) and ( worm_component > 0 ) then
				-- GamePrint("yes2")
				local dx = x - origin_x
				local dy = y - origin_y
				
				-- Normalize the direction vector
				local magnitude = math.sqrt(dx * dx + dy * dy)
				if magnitude > 0 then
					dx = dx / magnitude
					dy = dy / magnitude
				end
				
				-- Scale the normalized vector by the desired force
				local force_magnitude = 0.1 -- Adjust this value to set the strength of the push
				local force_x = dx * force_magnitude
				local force_y = dy * force_magnitude
				
				-- Apply the force to the worm's target vector
				local vx, vy = ComponentGetValue2(worm_component, "mTargetVec")
				ComponentSetValue2(worm_component, "mTargetVec", vx - force_x, vy - force_y)
			end
			
			local distance_from_origin = get_distance(x, y, origin_x, origin_y)
			if distance_from_origin < 200 then
				ComponentSetValue2( going_back_to_home_var_comp, "value_int", 0 )
			end
		-- elseif times_executed % 100 then
		
			-- EntitySetTransform(target_id, x + Random(-512, 512), y + Random(-512, 512))
			
			-- ComponentSetValue2( worm_ai_component, "mRandomTarget", origin_x, origin_y )
			-- ComponentSetValue2( worm_ai_component, "new_random_target_check_every", 9999 )
		
		end
		if ( BiomeMapGetName( x, y ) ~= BiomeMapGetName( origin_x, origin_y ) ) or ( y < 200 ) then
			ComponentSetValue2( going_back_to_home_var_comp, "value_int", 1 )
			
		end
		
		if ( #players > 0 ) and ( is_invisible(players[1]) == false ) then
			local px, py = EntityGetTransform(players[1])
			-- local distance_player = get_distance(x, y, px, py)
			if ( py > 200 ) then
				ComponentSetValue2( worm_ai_component, "mTargetEntityId", players[1] )
			end
		end
		
		-- ComponentSetValue2( worm_ai_component, "mTargetEntityId", target_id )
		-- ComponentSetValue2( worm_ai_component, "new_hunt_target_check_every", 1 )
	end
	
end