dofile_once( "data/scripts/lib/utilities.lua" )
dofile_once( "data/scripts/game_helpers.lua" )
dofile_once( "data/scripts/invisibility_check.lua" )

function collision_trigger(colliding_entity)

	local times = ComponentGetValue2(GetUpdatedComponentID(), "mTimesExecuted")

	local theEntity = GetUpdatedEntityID()
	local players = {}
	local player = EntityGetWithName( "DEBUG_NAME:player" )
	if ( player ~= nil ) and ( player > 0 ) then
		players = {player}
	else
		local polymorphed_player = EntityGetClosestWithTag(rx, ry, "polymorphed_player")
		if ( polymorphed_player ~= nil ) and ( polymorphed_player > 0 ) then
			players = {polymorphed_player}
		end
	end
	local player_x, player_y, player_scale_x = 0, 0, 1

	local x, y = EntityGetTransform( theEntity )

	local animal_ai_component = EntityGetFirstComponentIncludingDisabled(theEntity, "AnimalAIComponent")
	local physics_body_component = EntityGetFirstComponentIncludingDisabled(theEntity, "PhysicsBodyComponent")
	local sprite_component = EntityGetFirstComponentIncludingDisabled(theEntity, "SpriteComponent")
	local hitbox_component = EntityGetFirstComponentIncludingDisabled(theEntity, "HitboxComponent")

	local character_data_component = EntityGetFirstComponent(theEntity, "CharacterDataComponent")
	local vel_x, vel_y = ComponentGetValueVector2(character_data_component, "mVelocity")

	local physics_vel_x, physics_vel_y = PhysicsGetComponentVelocity( theEntity, physics_body_component )

	local number = GetValueNumber("number", 0)
	if number > 0 then
		SetValueNumber("number", number - 1)
		-- ComponentSetValue2(sprite_component, "rect_animation", "emerge")
		-- ComponentSetValue2(sprite_component, "next_rect_animation", "stand")
		if ComponentHasTag( sprite_component, "character" ) == true then
			ComponentRemoveTag( sprite_component, "character" )
		end
	elseif number <= 0 then
		if ComponentHasTag( sprite_component, "character" ) == false then
			ComponentAddTag( sprite_component, "character" )
		end
		SetValueNumber("number", 0)
	end

	function disable()

		if ComponentGetIsEnabled( physics_body_component ) == false then
			-- GamePlayAnimation( theEntity, "emerge", 1, "stand", 1 )
			GamePlaySound("mods/new_enemies/data/sound/more_sounds/my_mod_audio.bank", "snd_mod/bottle_pop", x, y)
			SetValueNumber("number", 30)
			EntityRefreshSprite( theEntity, sprite_component )
		end
		
		EntityRemoveComponent(theEntity, animal_ai_component)
		EntitySetComponentIsEnabled( theEntity, sprite_component, false )
		EntitySetComponentIsEnabled( theEntity, hitbox_component, false )
		EntitySetComponentIsEnabled( theEntity, physics_body_component, true )
		PhysicsApplyForce(theEntity, vel_x * 4 , vel_y * 4 )
		
	end

	function reenable()
	  if ( animal_ai_component == nil ) or ( animal_ai_component == 0 ) then
		GamePlaySound("mods/new_enemies/data/sound/more_sounds/my_mod_audio.bank", "snd_mod/bottle_pop", x, y)
		EntityAddComponent(theEntity, "AnimalAIComponent", {
				preferred_job="JobDefault",
				escape_if_damaged_probability="0",
				attack_melee_damage_min="0.4",
				attack_melee_damage_max="0.8",
				attack_melee_action_frame="2",
				attack_dash_enabled="1",
				attack_dash_distance="200",
				creature_detection_range_x="250",
				creature_detection_range_y="250",
				food_material="meat",
				needs_food="0",
				sense_creatures="1",
				attack_ranged_enabled="0",
				can_fly="0",
				defecates_and_pees="0",
		})
		
		EntitySetComponentIsEnabled( theEntity, sprite_component, true )
		if ComponentHasTag( sprite_component, "character" ) == true then
			ComponentRemoveTag( sprite_component, "character" )
		end
		ComponentSetValue2(sprite_component, "rect_animation", "emerge")
		ComponentSetValue2(sprite_component, "next_rect_animation", "stand")
		EntitySetComponentIsEnabled( theEntity, hitbox_component, true )
		EntitySetComponentIsEnabled( theEntity, physics_body_component, false )
		
		ComponentSetValue2(character_data_component, "mVelocity", physics_vel_x * 20, physics_vel_y * 20 )
		
	  end
	end


	if times < 2 then
		reenable()
		disable()
	end

	local has_player_tag = EntityHasTag(theEntity, "polymorphed_player")
	if ( has_player_tag == false ) then
		if #players > 0 and ( EntityGetFirstComponent( players[1], "GenomeDataComponent") ~= nil ) and ( EntityGetFirstComponent( players[1], "GenomeDataComponent") > 0 ) and ( EntityGetHerdRelation(theEntity, players[1]) < 40 ) and ( is_invisible(players[1]) == false ) then
			player_x, player_y, succ, player_scale_x = EntityGetTransform(players[1])

			
			local looking_at_crab = (player_x > x and player_scale_x < 0) or (player_x < x and player_scale_x > 0)

			if looking_at_crab and animal_ai_component then
			if math.abs(vel_x) <= 16 and math.abs(vel_y) <= 16 then
				disable()
			end
			
			elseif not looking_at_crab and not animal_ai_component then
				reenable()
			end
			
			local dist = get_distance(x, y, player_x, player_y)
			if dist <= 300 then
				ComponentSetValue2(GetUpdatedComponentID(), "execute_every_n_frame", 5)
			else
				ComponentSetValue2(GetUpdatedComponentID(), "execute_every_n_frame", 200)
			end
		end
	else
		local controls_component = EntityGetFirstComponentIncludingDisabled(theEntity, "ControlsComponent")
		local input = {
			ComponentGetValue(controls_component, "mButtonDownRight"), --right
			ComponentGetValue(controls_component, "mButtonDownUp"), --up
			ComponentGetValue(controls_component, "mButtonDownLeft"), --left
			ComponentGetValue(controls_component, "mButtonDownDown"), --down
		}

		if ( input[1] == "1" ) or ( input[2] == "1" ) or ( input[3] == "1" ) or ( input[4] == "1" ) then
			reenable()
		else
			if math.abs(vel_x) <= 16 and math.abs(vel_y) <= 16 then
				disable()
			end
		end
	end
end