dofile_once("data/scripts/lib/utilities.lua")

function approx_rolling_average(avg, new_sample, n)
  avg = avg - avg / n;
  avg = avg + new_sample / n;
  return avg
end

local old_average = GetValueNumber("average_rotation", 0)

function get_direction( x1, y1, x2, y2 )
  return math.atan2( ( y2 - y1 ), ( x2 - x1 ) )
end

function RadsToVec(rads)
    return math.cos(rads), math.sin(rads)
end

function rotDiff(to_angle, from_angle) 
    local end1 = math.pi
    local end2 = -end1

    if from_angle < end2 then 
        from_angle = end1 + (from_angle - end2)
    elseif from_angle > end1 then 
        from_angle = end2 + (from_angle - end1)
    end
    if to_angle < end2 then 
        to_angle = end1 + (to_angle - end2)
    elseif to_angle > end1 then 
        to_angle = end2 + (to_angle - end1)
    end


    local route1 = to_angle-from_angle
    local route2 = (end2-from_angle) + (to_angle-end1)
    local route3 = (end1-from_angle) + (to_angle-end2)

    if math.abs(route1) <= math.abs(route2) then
        if math.abs(route1) <= math.abs(route3) then
            return route1
        else
            return route3
        end
    else
        if math.abs(route2) <= math.abs(route3) then
            return route2
        else
            return route3
        end
    end
end--rotational difference
function rotatePoint(x,y,cx,cy,rads)
    local dist = get_distance(x,y,cx,cy)
    local angle = get_direction(cx,cy,x,y) + rads
    local add_x, add_y = RadsToVec(angle)
    local new_x, new_y = cx+(add_x*dist), cy+(add_y*dist)
    return new_x, new_y
end--> rotated x, y

local entity_id = GetUpdatedEntityID()
local x, y, rot, scale_x = EntityGetTransform(entity_id)

--Moving our normal detectors based on entity rotation
local right_corner = { x1 = -4, y1 = 0 }
local right_x, right_y = vec_rotate(right_corner.x1, right_corner.y1, rot)
right_x = x + right_x
right_y = y + right_y

local left_corner = { x2 = 4, y2 = 0 }
local left_x, left_y = vec_rotate(left_corner.x2, left_corner.y2, rot)
left_x = x + left_x
left_y = y + left_y

local right_corner2 = { x1 = -16, y1 = 0 }
local right_x2, right_y2 = vec_rotate(right_corner2.x1, right_corner2.y1, rot)
right_x2 = x + right_x2
right_y2 = y + right_y2

local left_corner2 = { x2 = 16, y2 = 0 }
local left_x2, left_y2 = vec_rotate(left_corner2.x2, left_corner2.y2, rot)
left_x2 = x + left_x2
left_y2 = y + left_y2

function GetSurfaceNormal(pos_x, pos_y, ray_length, ray_count)
	local total_normal_x = 0
	local total_normal_y = 0
	local total_distance = 0
	local found_normal = false

	for i = 1, ray_count do
		local angle = (i - 1) * (2 * math.pi / ray_count)
		local ray_x = pos_x + ray_length * math.cos(angle)
		local ray_y = pos_y + ray_length * math.sin(angle)

		local did_hit, hit_x, hit_y = RaytracePlatforms(pos_x, pos_y, ray_x, ray_y)

		if did_hit then
			found_normal = true
			local normal_x, normal_y = hit_x - pos_x, hit_y - pos_y
			total_normal_x = total_normal_x + normal_x
			total_normal_y = total_normal_y + normal_y
			total_distance = total_distance + math.sqrt(normal_x^2 + normal_y^2)
		end
	end

	if found_normal then
		local average_normal_x = total_normal_x / ray_count
		local average_normal_y = total_normal_y / ray_count
		local average_distance = total_distance / ray_count
		return true, average_normal_x, average_normal_y, average_distance
	else
		return false, 0, 0, 0
	end
end

local found_normal, normal_x, normal_y, approximate_distance_from_surface = GetSurfaceNormal(right_x, right_y, 10, 16)
local found_normal2, normal_x2, normal_y2, approximate_distance_from_surface = GetSurfaceNormal(left_x, left_y, 10, 16)
local found_normal3, normal_x3, normal_y3, approximate_distance_from_surface = GetSurfaceNormal(x, y, 10, 16)

local found_normal4, normal_x4, normal_y4, approximate_distance_from_surface = GetSurfaceNormal(right_x2, right_y2, 10, 16)
local found_normal5, normal_x5, normal_y5, approximate_distance_from_surface = GetSurfaceNormal(left_x2, left_y2, 10, 16)

local normal_in_solid, hit_x, hit_y = RaytracePlatforms( right_x, right_y, right_x, right_y )
local normal_in_solid2, hit_x2, hit_y2 = RaytracePlatforms( left_x, left_y, left_x, left_y )
local normal_in_solid3, hit_x3, hit_y3 = RaytracePlatforms( x, y, x, y )
local normal_in_solid4, hit_x4, hit_y4 = RaytracePlatforms( right_x2, right_y2, right_x2, right_y2 )
local normal_in_solid5, hit_x5, hit_y5 = RaytracePlatforms( left_x2, left_y2, left_x2, left_y2 )

normal_count = 5

if normal_in_solid or not found_normal then
	normal_x = 0
	normal_y = 0
	normal_count = normal_count - 1
end

if normal_in_solid2 or not found_normal2 then
	normal_x2 = 0
	normal_y2 = 0
	normal_count = normal_count - 1
end

if normal_in_solid3 or not found_normal3 then
	normal_x3 = 0
	normal_y3 = 0
	normal_count = normal_count - 1
end

if normal_in_solid4 or not found_normal4 then
	normal_x4 = 0
	normal_y4 = 0
	normal_count = normal_count - 1
end

if normal_in_solid5 or not found_normal5 then
	normal_x5 = 0
	normal_y5 = 0
	normal_count = normal_count - 1
end


if normal_count > 0 then

	local normal_x_average = (normal_x + normal_x2 + normal_x3 + normal_x4 + normal_x5) / normal_count
	local normal_y_average = (normal_y + normal_y2 + normal_y3 + normal_y4 + normal_y5) / normal_count

	local dir = get_direction(0, 0, normal_x_average, normal_y_average)
	local degree_shift = math.rad(-90) + dir -- 90 degree rotation to properly orient the entity
	local new_average = approx_rolling_average(old_average, degree_shift, 5) -- if the number is higher than 5 the turning speed would be smoother and slower, if less it would be more instantaneous
	SetValueNumber("average_rotation", new_average)

	local max_angle = math.rad(45)
	local new_average_capped_angle = math.min(max_angle, math.max(new_average, -max_angle))
	EntitySetTransform(entity_id, x, y, new_average_capped_angle)
end


for i, child_id in ipairs(EntityGetAllChildren(entity_id) or {}) do
	if EntityGetName(child_id) == "hand" then
		-- Get the world coordinates of the troot
		troot_x, troot_y = EntityGetTransform(child_id)
		
		-- Get the world coordinates of the hand hotspot
		local hx, hy = EntityGetHotspot(entity_id, "hand", true, true)
		
		-- Get the InheritTransformComponent of the child entity
		local inheritcomp = EntityGetFirstComponent(child_id, "InheritTransformComponent")
		
		-- Get the current transform values of the component
		local tx, ty, scale_x_local, scale_y2, rot2 = ComponentGetValue2(inheritcomp, "Transform")
		
		-- Calculate the relative differences
		local x_diff = ( hx - x ) * scale_x
		local y_diff = hy - y
		
		-- Set the relative differences in the InheritTransformComponent
		ComponentSetValue2(inheritcomp, "Transform", x_diff, y_diff, scale_x_local, scale_y2, rot2)
	end
end

--Constrain trunk segments
function constrain_and_elongate(entity, x2, y2, max_dist, y_scale)
	local projcomp = EntityGetFirstComponent(entity, "VelocityComponent")
	local proj_vel_x, proj_vel_y = ComponentGetValue2(projcomp, "mVelocity")
		
	local x1, y1 = EntityGetTransform(entity)
	
	--Control Neck limb

	local angle = get_direction(x1, y1, x2, y2)
	local distance = get_distance(x1, y1, x2, y2)

	local sprite_width = 10
	local scale = math.max(0, distance) / sprite_width

	EntitySetTransform(entity, x1, y1, angle, scale, y_scale)

	if (not x1) or (not x2) then return nil end
	
	local dist = math.sqrt(((x2-x1)^2) + ((y2-y1)^2))
	local target_x = x2 + (((x1-x2)/dist)*max_dist)
	local target_y = y2 + (((y1-y2)/dist)*max_dist)
	local target_vel_x = target_x - x1
	local target_vel_y = target_y - y1
	
		
	  local dx = x1 - x2
	  local dy = y1 - y2
	  local dist = math.sqrt(dx*dx + dy*dy)
	  if ( dist > max_dist ) or ( dist < max_dist ) and (dist~=0) then
		x1 = x2 + (dx/dist) * max_dist
		y1 = y2 + (dy/dist) * max_dist
	  end

	if dist > max_dist then
		if proj_vel_x then
			ComponentSetValue2(projcomp, "mVelocity", proj_vel_x+(target_vel_x*5), proj_vel_y+(target_vel_y*5))
			EntitySetTransform(entity, x1, y1)
		end
	end
end

--Constrain body to a ring around the entity
function constrain_to_ring2(entity, x2, y2, max_dist, scale_of_y)
  local x1, y1, bla, blasx = EntityGetTransform(entity)
	--Control Neck limb

	local angle = get_direction(x1, y1, x2, y2)
	local distance = get_distance(x1, y1, x2, y2)

	local sprite_width = 10
	local scale = math.max(0, distance) / sprite_width
	
  local angle = get_direction(x1, y1, x2, y2)
  local dx = x1 - x2
  local dy = y1 - y2
  local dist = math.sqrt(dx*dx + dy*dy)
  if ( dist > max_dist ) or ( dist < max_dist ) and (dist~=0) then
    x1 = x2 + (dx/dist) * max_dist
    y1 = y2 + (dy/dist) * max_dist
  end
  EntitySetTransform(entity, x1, y1, angle, scale, scale_of_y)
end

--Constrain body to a ring around the entity
function constrain_to_ring(entity, x2, y2, max_dist, scale_of_y)
  local x1, y1, bla, blasx = EntityGetTransform(entity)
  local angle = get_direction(x1, y1, x2, y2)
  local dx = x1 - x2
  local dy = y1 - y2
  local dist = math.sqrt(dx*dx + dy*dy)
  if ( dist > max_dist ) or ( dist < max_dist ) and (dist~=0) then
    x1 = x2 + (dx/dist) * max_dist
    y1 = y2 + (dy/dist) * max_dist
  end
  if x1 < x2 then
	scale_calc = 1
  else
	scale_calc = -1
  end
  EntitySetTransform(entity, x1, y1, angle, blasx, scale_calc)
end

local ivalue = 0

local times_executed = ComponentGetValue2(GetUpdatedComponentID(), "mTimesExecuted")

local animalcomp = EntityGetFirstComponent(entity_id, "AnimalAIComponent")
local greatest_prey_id = ComponentGetValue2(animalcomp, "mGreatestPrey")
local players = {}
local player = EntityGetWithName( "DEBUG_NAME:player" )
if ( player ~= nil ) and ( player > 0 ) then
	players = {player}
else
	local polymorphed_player = EntityGetClosestWithTag(x, y, "polymorphed_player")
	if ( polymorphed_player ~= nil ) and ( polymorphed_player > 0 ) then
		players = {polymorphed_player}
	end
end
player = players[1]
local gprx, gpry  = EntityGetTransform(entity_id)
if greatest_prey_id > 0 then
	gprx, gpry  = EntityGetTransform(greatest_prey_id)
elseif player > 0 then
	gprx, gpry  = EntityGetTransform(player)
else
	gprx, gpry  = EntityGetTransform(entity_id)
end

local old_enemy_x = GetValueNumber("enemy_x", gprx)
local old_enemy_y = GetValueNumber("enemy_y", gpry)
if old_enemy_x and old_enemy_y and gprx and gpry then
	enemy_x = approx_rolling_average(old_enemy_x, gprx, 5)
	enemy_y = approx_rolling_average(old_enemy_y, gpry, 5)
else
	return
end
SetValueNumber("enemy_x", enemy_x)
SetValueNumber("enemy_y", enemy_y)

local animal_ai_component = EntityGetFirstComponentIncludingDisabled(entity_id, "AnimalAIComponent")
local state = ComponentGetValue2(animal_ai_component, "ai_state")
local difference = math.abs(13 - state)

local melee_delay = GetValueNumber("melee_delay", 0)
local ring_dist = GetValueNumber("ring_dist", 12)

if ( difference < 0.0001 ) then
	is_melee_attacking = true
	SetValueNumber("melee_delay", 60)
	if ring_dist > 0 then
		SetValueNumber("ring_dist", ring_dist - 0.5)
	end
elseif melee_delay <= 0 then
	is_melee_attacking = false
end

if melee_delay > 0 then
	SetValueNumber("melee_delay", melee_delay - 1)
end

if ( ring_dist < 12 ) and ( is_melee_attacking == false ) then
	SetValueNumber("ring_dist", ring_dist + 0.1)
end

--Getting all my children
local prevchild
for i, child in ipairs(EntityGetAllChildren(entity_id) or {}) do
  if EntityGetName(child) == "trunk" then
	ivalue = ivalue + 1
    local amount_of_children = 8
	local real_amount_of_children = 8
	y_value = 1

    if times_executed < 5 then
        EntitySetTransform(child, x, y)
    elseif times_executed < 10 then
        EntitySetComponentsWithTagEnabled(child, "enabled_in_world", true)
    end
	if times_executed > 20 then
		if ( is_melee_attacking == true ) and ( ring_dist <= 0 ) then
			EntitySetComponentsWithTagEnabled(child, "enabled_in_world", false)
		elseif is_melee_attacking == false then
			EntitySetComponentsWithTagEnabled(child, "enabled_in_world", true)
		end
	end
    if not prevchild then
		local segment_x, segment_y, segment_rot = EntityGetTransform(child)
        -- constrain_and_elongate(child, troot_x, troot_y, 14, y_value)
		constrain_to_ring2(child, troot_x, troot_y, ring_dist, y_value)
		local cx, cy, crot, cscale_x = EntityGetTransform(child)
		EntitySetTransform(child, troot_x, troot_y)
		
        --stuff Thatrius added
        local ratio1 = ivalue / real_amount_of_children
        local ratio2 = (real_amount_of_children - ivalue) / real_amount_of_children
        local angle_outward = get_direction(0, 0, scale_x, 0)
        local angle_aim = get_direction(enemy_x, enemy_y, segment_x, segment_y)
        local torque = (rotDiff(angle_outward, segment_rot)*ratio2) + (rotDiff(angle_aim, segment_rot)*ratio1)
        local targ_x, targ_y = rotatePoint(segment_x, segment_y, troot_x, troot_y, torque)
		
		local projcomp = EntityGetFirstComponent(child, "VelocityComponent")
		local proj_vel_x, proj_vel_y = ComponentGetValue2(projcomp, "mVelocity")
		ComponentSetValue2(projcomp, "mVelocity", ( proj_vel_x * 0.5 ) + ((targ_x-segment_x) * 9), ( proj_vel_y * 0.5 ) + ((targ_y-segment_y) * 9))
    else
        local segment_x, segment_y, segment_rot = EntityGetTransform(child)
		local prevsegmentx, prevsegmenty = EntityGetTransform(prevchild)
        -- constrain_and_elongate(child, prevsegmentx, prevsegmenty, 14, y_value)
		constrain_to_ring2(child, prevsegmentx, prevsegmenty, ring_dist * 0.82, y_value)

        --stuff Thatrius added
        local ratio1 = ivalue / real_amount_of_children
        local ratio2 = (real_amount_of_children - ivalue) / real_amount_of_children
        local angle_outward = get_direction(0, 0, scale_x, 0)
        local angle_aim = get_direction(enemy_x, enemy_y, segment_x, segment_y)
        local torque = (rotDiff(angle_outward, segment_rot)*ratio2) + (rotDiff(angle_aim, segment_rot)*ratio1)
        local targ_x, targ_y = rotatePoint(segment_x, segment_y, prevsegmentx, prevsegmenty, torque)
		
		local projcomp = EntityGetFirstComponent(child, "VelocityComponent")
		local proj_vel_x, proj_vel_y = ComponentGetValue2(projcomp, "mVelocity")
		ComponentSetValue2(projcomp, "mVelocity", ( proj_vel_x * 0.3 ) + ((targ_x-segment_x) * 10), ( proj_vel_y * 0.3 ) + ((targ_y-segment_y) * 10))
    end
    prevchild = child
	
	local trunk_end = EntityGetAllChildren(entity_id)[4]
	local segmentx, segmenty = EntityGetTransform(trunk_end)
	local targ_x, targ_y = x+GetValueNumber("trunk_end_x", x), y+GetValueNumber("trunk_end_y", y)
	if math.random(100)==1 then 
		SetValueNumber("trunk_end_x", math.random(50))
		SetValueNumber("trunk_end_y", math.random(50))
		local projcomp = EntityGetFirstComponent(trunk_end, "VelocityComponent")
		local proj_vel_x, proj_vel_y = ComponentGetValue2(projcomp, "mVelocity")
		ComponentSetValue2(projcomp, "mVelocity", proj_vel_x+(targ_x-segmentx), proj_vel_y+(targ_y-segmenty))
	end
  end
end