dofile_once("data/scripts/lib/utilities.lua")

function approx_rolling_average(avg, new_sample, n)
  avg = avg - avg / n;
  avg = avg + new_sample / n;
  return avg
end
function get_distance( x1, y1, x2, y2 )
	local result = math.sqrt( ( x2 - x1 ) ^ 2 + ( y2 - y1 ) ^ 2 )
	return result
end

function get_direction( x1, y1, x2, y2 )
	return math.atan2( ( y2 - y1 ), ( x2 - x1 ) )
end

function RadsToVec(rads)
    return math.cos(rads), math.sin(rads)
end

function rotDiff(to_angle, from_angle) 
    local end1 = math.pi
    local end2 = -end1

    if from_angle < end2 then 
        from_angle = end1 + (from_angle - end2)
    elseif from_angle > end1 then 
        from_angle = end2 + (from_angle - end1)
    end
    if to_angle < end2 then 
        to_angle = end1 + (to_angle - end2)
    elseif to_angle > end1 then 
        to_angle = end2 + (to_angle - end1)
    end


    local route1 = to_angle-from_angle
    local route2 = (end2-from_angle) + (to_angle-end1)
    local route3 = (end1-from_angle) + (to_angle-end2)

    if math.abs(route1) <= math.abs(route2) then
        if math.abs(route1) <= math.abs(route3) then
            return route1
        else
            return route3
        end
    else
        if math.abs(route2) <= math.abs(route3) then
            return route2
        else
            return route3
        end
    end
end
function rotatePoint(x,y,cx,cy,rads)
    local dist = get_distance(x,y,cx,cy)
    local angle = get_direction(cx,cy,x,y) + rads
    local add_x, add_y = RadsToVec(angle)
    local new_x, new_y = cx+(add_x*dist), cy+(add_y*dist)
    return new_x, new_y
end

local entity_id = GetUpdatedEntityID()
local root_id = EntityGetParent( entity_id )
EntitySetComponentsWithTagEnabled(root_id, "enabled_in_world", false)
-- local rx, ry  = EntityGetTransform(root_id)
local rx, ry = EntityGetHotspot( root_id, "hand", true, false )

local x, y, rot, scale_x  = EntityGetTransform(entity_id)

local times = ComponentGetValue2(GetUpdatedComponentID(), "mTimesExecuted")

SetRandomSeed( GameGetFrameNum(), x + y + entity_id )

--Constrain trunk segments
function constrain_and_elongate(entity, x2, y2, max_dist)
	local projcomp = EntityGetFirstComponent(entity, "VelocityComponent")
	local proj_vel_x, proj_vel_y = ComponentGetValue2(projcomp, "mVelocity")
		
	local x1, y1 = EntityGetTransform(entity)
	
	--Control Neck limb

	local angle = get_direction(x1, y1, x2, y2)
	local distance = get_distance(x1, y1, x2, y2)

	local sprite_width = 1
	local scale = math.max(0, distance) / sprite_width

	EntitySetTransform(entity, x1, y1, angle, scale)

	if (not x1) or (not x2) then return nil end
	
	local dist = math.sqrt(((x2-x1)^2) + ((y2-y1)^2))
	local target_x = x2 + (((x1-x2)/dist)*max_dist)
	local target_y = y2 + (((y1-y2)/dist)*max_dist)
	local target_vel_x = target_x - x1
	local target_vel_y = target_y - y1
	
		
	  local dx = x1 - x2
	  local dy = y1 - y2
	  local dist = math.sqrt(dx*dx + dy*dy)
	  if ( dist > max_dist ) or ( dist < max_dist ) and (dist~=0) then
		x1 = x2 + (dx/dist) * max_dist
		y1 = y2 + (dy/dist) * max_dist
	  end

	if dist > max_dist then
		if proj_vel_x then
			ComponentSetValue2(projcomp, "mVelocity", ( proj_vel_x + target_vel_x ) * 0.9, ( proj_vel_y + target_vel_y ) * 0.9 )
		end
	end
end

function constrain_and_elongate2(entity, prev_child, max_dist)
	local projcomp = EntityGetFirstComponent(entity, "VelocityComponent")
	local proj_vel_x, proj_vel_y = ComponentGetValue2(projcomp, "mVelocity")
	
	local x2, y2 = EntityGetTransform(prev_child)
		
	local x1, y1 = EntityGetTransform(entity)
	
	--Control Neck limb

	local angle = get_direction(x1, y1, x2, y2)
	local distance = get_distance(x1, y1, x2, y2)

	local sprite_width = 1
	local scale = math.max(0, distance) / sprite_width

	EntitySetTransform(entity, x1, y1, angle, scale)

	if (not x1) or (not x2) then return nil end
	
	local dist = math.sqrt(((x2-x1)^2) + ((y2-y1)^2))
	local target_x = x2 + (((x1-x2)/dist)*max_dist)
	local target_y = y2 + (((y1-y2)/dist)*max_dist)
	local target_vel_x = target_x - x1
	local target_vel_y = target_y - y1
	

	local dx = x1 - x2
	local dy = y1 - y2
	local dist = math.sqrt(dx*dx + dy*dy)
	if ( dist > max_dist ) or ( dist < max_dist ) and (dist~=0) then
		x1 = x2 + (dx/dist) * max_dist
		y1 = y2 + (dy/dist) * max_dist
	end
	
	local projcomp2 = EntityGetFirstComponent(prev_child, "VelocityComponent")
	local proj_vel_x2, proj_vel_y2 = ComponentGetValue2(projcomp2, "mVelocity")
	local multiplier = 10
	if dist > max_dist then
		if proj_vel_x then
			ComponentSetValue2(projcomp, "mVelocity", ( proj_vel_x + target_vel_x * multiplier ) * 0.95, 4 + ( proj_vel_y + target_vel_y * multiplier ) * 0.95 )
			ComponentSetValue2(projcomp2, "mVelocity", ( proj_vel_x2 - target_vel_x * multiplier ), ( proj_vel_y2 - target_vel_y * multiplier ) )
		end
	end
end

function constrain_force(entity, prev_child, max_dist)
	local projcomp = EntityGetFirstComponent(entity, "VelocityComponent")
	local proj_vel_x, proj_vel_y = ComponentGetValue2(projcomp, "mVelocity")
	
	local x2, y2 = EntityGetTransform(prev_child)
		
	local x1, y1 = EntityGetTransform(entity)

	if (not x1) or (not x2) then return nil end
	
	local dist = math.sqrt(((x2-x1)^2) + ((y2-y1)^2))
	local target_x = x2 + (((x1-x2)/dist)*max_dist)
	local target_y = y2 + (((y1-y2)/dist)*max_dist)
	local target_vel_x = target_x - x1
	local target_vel_y = target_y - y1
	

	local dx = x1 - x2
	local dy = y1 - y2
	local dist = math.sqrt(dx*dx + dy*dy)
	if ( dist > max_dist ) or ( dist < max_dist ) and (dist~=0) then
		x1 = x2 + (dx/dist) * max_dist
		y1 = y2 + (dy/dist) * max_dist
	end
	
	local projcomp2 = EntityGetFirstComponent(prev_child, "VelocityComponent")
	local proj_vel_x2, proj_vel_y2 = ComponentGetValue2(projcomp2, "mVelocity")
	local multiplier = 10
	if dist > max_dist then
		if proj_vel_x then
			ComponentSetValue2(projcomp, "mVelocity", proj_vel_x + ( target_vel_x * 0.01 ) , proj_vel_y + ( target_vel_y * 0.01 ) )
		end
	end
end

function retract(entity, x2, y2, max_dist)
	local projcomp = EntityGetFirstComponent(entity, "VelocityComponent")
	local proj_vel_x, proj_vel_y = ComponentGetValue2(projcomp, "mVelocity")
	local x1, y1 = EntityGetTransform(entity)
	if (not x1) or (not x2) then return nil end
		
	local dist = math.sqrt(((x2-x1)^2) + ((y2-y1)^2))
	local target_x = x2 + (((x1-x2)/dist)*max_dist)
	local target_y = y2 + (((y1-y2)/dist)*max_dist)
	local target_vel_x = target_x - x1
	local target_vel_y = target_y - y1

	if dist > max_dist then
		if proj_vel_x then
			ComponentSetValue2(projcomp, "mVelocity", proj_vel_x+(target_vel_x*100), proj_vel_y+(target_vel_y*100))
			EntitySetTransform(entity, target_x, target_y)
		end
	end
end

--Getting all my children
local prevchild
local children = EntityGetAllChildren(entity_id)
for i, child in ipairs(children or {}) do
	if EntityGetName(child) == "trunk" then
		local cx, cy = EntityGetTransform(child)
		-- GameCreateSpriteForXFrames( "data/projectiles_gfx/radiobot_radioactivity.png", cx, cy, true, 0, 0, 1, true )
		
		if times < 5 then
			EntitySetTransform(child, x, y)
		elseif times < 10 then
			EntitySetComponentsWithTagEnabled(child, "enabled_in_world", true)
		end
		if not prevchild then
			local segment_x, segment_y, segment_rot = EntityGetTransform(child)
			-- constrain_and_elongate(child, troot_x, troot_y, 14, y_value)
			constrain_and_elongate(child, rx, ry, 5)
		elseif child == children[#children] then
			constrain_force(entity_id, children[( #children ) - 1], 5)
			constrain_and_elongate2(child, prevchild, 5)
		else
			local segment_x, segment_y, segment_rot = EntityGetTransform(child)
			constrain_and_elongate2(child, prevchild, 5)
		end
		
		prevchild = child
	end
end

if times > 60 then
	if ( children[( #children ) - 1] ~= nil ) and ( children[( #children ) - 1] > 0 ) then
		local varcomp = EntityGetFirstComponent(children[( #children ) - 1], "VariableStorageComponent")
		local init_bool = ComponentGetValue2(varcomp, "value_bool")
		if init_bool == false then
			local x2, y2 = EntityGetTransform(children[( #children ) - 1])
			local dist = get_distance( x, y, x2, y2 )
			ComponentSetValue2( varcomp, "value_float", dist )
			ComponentSetValue2( varcomp, "value_bool", true )
		end
		local constrain_dist = ComponentGetValue2(varcomp, "value_float")
		if constrain_dist > 0 then
			local x2, y2 = EntityGetTransform(children[( #children ) - 1])
			retract(entity_id, x2, y2, constrain_dist)
			ComponentSetValue2( varcomp, "value_float", constrain_dist - 1 )
			local dist = get_distance( x, y, x2, y2 )
			if dist <= 3 then
				EntityKill( children[ ( #children ) - 1 ] )
			end
		else
			ComponentSetValue2( varcomp, "value_float", 30 )
		end
	else
		local varcomp = EntityGetFirstComponent(root_id, "VariableStorageComponent")
		local init_bool = ComponentGetValue2(varcomp, "value_bool")
		if init_bool == false then
			local dist = get_distance( x, y, rx, ry )
			ComponentSetValue2( varcomp, "value_float", dist )
			ComponentSetValue2( varcomp, "value_bool", true )
		end
		local constrain_dist = ComponentGetValue2(varcomp, "value_float")
		if constrain_dist > 0 then
			retract(entity_id, rx, ry, constrain_dist)
			ComponentSetValue2( varcomp, "value_float", constrain_dist - 1 )
			local dist = get_distance( x, y, rx, ry )
			if dist <= 3 then
				EntityKill(entity_id)
				EntitySetComponentsWithTagEnabled(root_id, "enabled_in_world", true)
				
				--animation
				local sprite_comps = EntityGetComponent(root_id, "SpriteComponent") or {}
				for i, comp in ipairs(sprite_comps) do
					EntityRefreshSprite( root_id, comp )
				end
			end
		else
			ComponentSetValue2( varcomp, "value_float", 30 )
		end
	end
end