dofile_once("data/scripts/lib/utilities.lua")

function collision_trigger(colliding_entity_id)
	-- Utility function to split a string by delimiter and return a table
	function split(str, delimiter)
		local result = {}
		for match in (str..delimiter):gmatch("(.-)"..delimiter) do
			table.insert(result, match)
		end
		return result
	end

	-- Utility function to join a table into a string with a delimiter
	function join(tbl, delimiter)
		return table.concat(tbl, delimiter)
	end
	
    local entity_id = GetUpdatedEntityID()
    local pos_x, pos_y = EntityGetTransform(entity_id)

    -- Retrieve the global value storing the spawned worlds
    local spawned_worlds = GlobalsGetValue("new_enemies_hellion_spawned", "")

    -- Convert the global value to a table
    local spawned_worlds_table = split(spawned_worlds, ",")

    -- Function to check if a value exists in a table
    function table_contains(tbl, val)
        for _, v in ipairs(tbl) do
            if v == val then
                return true
            end
        end
        return false
    end

    -- Get the current parallel world index
    local current_world_x, current_world_y = GetParallelWorldPosition(pos_x, pos_y)
    local current_world_index = tostring(current_world_x) .. ","
	
	local current_world_occupied = table_contains(spawned_worlds_table, tostring(current_world_x))

    -- Check if the boss has already been spawned in the current world
    if current_world_occupied == true then
        EntityKill(entity_id)
	elseif current_world_occupied == false then
		-- Add the current world index to the table and update the global value
		table.insert(spawned_worlds_table, current_world_index)
		GlobalsSetValue("new_enemies_hellion_spawned", join(spawned_worlds_table, ","))

		-- Spawn the boss and kill the spawner entity
		EntityLoad("data/entities/animals/hellion.xml", pos_x, pos_y)
		EntityKill(entity_id)
    end
end