dofile_once("data/scripts/lib/utilities.lua")

function shot(projectile_id)
	local this_lua = GetUpdatedComponentID()

	local entity_id = GetUpdatedEntityID()
	local x, y = EntityGetTransform(entity_id)
    local px, py = EntityGetTransform(projectile_id)
    local vx, vy = GameGetVelocityCompVelocity(projectile_id)

    --Calculate the magnitude of the velocity vector
    local magnitude = math.sqrt(vx * vx + vy * vy)
	
	if magnitude == 0 then
		local relative_x = x - px
		local relative_y = ( y - 5 ) - py
		magnitude = math.sqrt(relative_x * relative_x + relative_y * relative_y)
		
		local velocity_multiplier = -1
		vx = relative_x * velocity_multiplier
		vy = relative_y * velocity_multiplier
	end

    -- If magnitude is not zero, normalize the velocity vector
    if magnitude ~= 0 then
        vx = vx / magnitude
        vy = vy / magnitude
	else
    end

    -- Multiply the normalized velocity by the desired magnitude (e.g., 100)
    local hook_velocity = 500
    vx = vx * hook_velocity
    vy = vy * hook_velocity

	function shoot_projectile_from_projectile( who_shot, entity_file, x, y, vel_x, vel_y )
		local entity_id = EntityLoad( entity_file, x, y )
		local herd_id   = get_herd_id( who_shot )
		local entity_that_shot = GetUpdatedEntityID()

		local who_shot_creature = 0
		edit_component( who_shot, "ProjectileComponent", function(comp,vars)
			who_shot_creature = ComponentGetValue2( comp, "mWhoShot" )
			ComponentSetValue2( comp, "mEntityThatShot", entity_that_shot )
		end)


		GameShootProjectile( who_shot_creature, x, y, x+vel_x, y+vel_y, entity_id, false, who_shot )

		edit_component( entity_id, "ProjectileComponent", function(comp,vars)
			vars.mWhoShot       = component_get_value_int( who_shot, "ProjectileComponent", "mWhoShot", 0 )
			vars.mShooterHerdId = component_get_value_int( who_shot, "ProjectileComponent", "mShooterHerdId", 0 )
		end)

		edit_component( entity_id, "VelocityComponent", function(comp,vars)
			ComponentSetValueVector2( comp, "mVelocity", vel_x, vel_y)
		end)
		
		if EntityHasTag( who_shot, "friendly_fire_enabled" ) then
			EntityAddTag( entity_id, "friendly_fire_enabled" )
			
			edit_component( entity_id, "ProjectileComponent", function(comp,vars)
				ComponentSetValue2( comp, "friendly_fire", true )
				ComponentSetValue2( comp, "collide_with_shooter_frames", 6 )
			end)
		end

		return entity_id
	end

	function remove_status_effects( game_effect_entity )
		local components = EntityGetComponentIncludingDisabled( game_effect_entity, "UIIconComponent", "enabled_in_world"  )
		if ( components ~= nil ) then
			for _, comp_id in pairs(components) do
				EntityRemoveComponent( game_effect_entity, comp_id )
			end
		end
	end
	
	function add_status_effect( game_effect_entity, icon_file, name, description )
		if ( game_effect_entity ~= nil ) and ( game_effect_entity > 0 ) then
			
			local components = EntityGetComponentIncludingDisabled( game_effect_entity, "UIIconComponent", "enabled_in_world" )
			if ( components == nil ) or ( #components < 1 ) then
				local newcomp = EntityAddComponent2( game_effect_entity, "UIIconComponent",
				{
					name = name,
					description = description,
					icon_sprite_file = icon_file,
					display_above_head = true,
					display_in_hud = true,
					is_perk = false,
				})
				ComponentAddTag( newcomp, "enabled_in_world" )
			end
		end
	end
	
	local path = "data/entities/projectiles/deck/hook_mwyah.xml"
	
	local mwyah = EntityGetWithName("$mwyah_phase1")
	local mwyah2 = EntityGetWithName("$mwyah")
	if ( mwyah ~= nil ) and  ( mwyah > 0 ) then
		--Getting the arena checker
		local children = EntityGetAllChildren( mwyah, "pitcheck_b" )
		local arena_center_id = children[1]
		
		local var_comp_terrain = get_variable_storage_component( arena_center_id, "is_clearing_terrain" )
		local current_state = ComponentGetValue2(var_comp_terrain, "value_int")
		local state_prev_var_comp = get_variable_storage_component( arena_center_id, "old_state" )
		local old_state = ComponentGetValue2(state_prev_var_comp, "value_int")
		ComponentSetValue2(state_prev_var_comp, "value_int", current_state)
		
		if old_state ~= current_state then
			remove_status_effects( entity_id )
		end

		if current_state == 4 then
			SetValueInteger("init_frame", GameGetFrameNum())
		end
		
		if current_state == 5 then
			path = "data/entities/projectiles/deck/teleport_mwyah.xml"
			add_status_effect( entity_id, "data/ui_gfx/status_indicators/telebolt_effect.png", "$telebolts_status_name", "$telebolts_status_desc" )
		elseif current_state == 6 then
			path = "data/entities/projectiles/deck/long_distance_cast_mwyah.xml"
			add_status_effect( entity_id, "data/ui_gfx/status_indicators/long_distances_effect.png", "$longdistancetelebolts_status_name", "$longdistancetelebolts_status_desc" )
		elseif current_state == 7 then
			path = "data/entities/projectiles/deck/mist_mwyah.xml"
			add_status_effect( entity_id, "data/ui_gfx/status_indicators/mists_effect.png", "$bloodmist_status_name", "$bloodmist_status_desc" )
			vx = 0
			vy = 0
		elseif current_state == 8 then
			path = "data/entities/projectiles/deck/swapper_mwyah.xml"
			add_status_effect( entity_id, "data/ui_gfx/status_indicators/swappers_effect.png", "$swapperbolts_status_name", "$swapperbolts_status_desc" )
			vx = vx * 1.5
			vy = vy * 1.5
		elseif current_state == 3 then
			add_status_effect( entity_id, "data/ui_gfx/status_indicators/hooks_effect.png", "$hooks_status_name", "$hooks_status_desc" )
		end
		
		-- Shoot the projectile with the modified velocity
		local init_frame = GetValueInteger("init_frame", GameGetFrameNum())
		local now = GameGetFrameNum()
		local difference = now - init_frame
		if ( current_state ~= 14 ) and ( current_state ~= 2 ) then
			if ( difference > 175 ) and ( current_state ~= 13 ) then
				if ( EntityHasTag( projectile_id, "projectile_not" ) == false ) then
					shoot_projectile_from_projectile(projectile_id, path, px, py, vx, vy)
				end
			else
				add_status_effect( entity_id, "data/ui_gfx/status_indicators/neutralized.png", "$effect_neutralized", "$effectdesc_neutralized" )
			end

			EntityApplyTransform(projectile_id, 66666666, 66666666)
			
			local all_comps = EntityGetAllComponents( projectile_id ) or {}
			for i, comp in ipairs(all_comps) do
				EntityRemoveComponent( projectile_id, comp )
			end
			
			local lua_comp = EntityAddComponent2(projectile_id, "LuaComponent", {
				script_source_file="data/scripts/animals/kill_entity.lua",
				execute_every_n_frame=1,
				execute_on_added=false
			})
			-- EntityKill(projectile_id)
		else
			function split_string(inputstr, sep)
			  sep = sep or "%s"
			  local t= {}
			  for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
				table.insert(t, str)
			  end
			  return t
			end

			function toboolean(str)
				return str == "1"
			end

			local character_platforming_component = EntityGetFirstComponentIncludingDisabled(entity_id, "CharacterPlatformingComponent")
			local var_store = get_variable_storage_component(entity_id, "new_enemies_player_mwyah_phase2_old_values")
			if ( var_store ~= nil ) and ( var_store > 0 ) then
				local stored_values = ComponentGetValue2(var_store, "value_string")
				local properties = split_string(stored_values, ",")
				
				for i, v in ipairs(properties) do
					local name, prop_type, value = string.match(v, "(.+):(.+)=(.+)")
					value = _G["to"..prop_type](value)
					ComponentSetValue2(character_platforming_component, name, value)
				end
				
				EntityRemoveComponent(entity_id, var_store)
				
			end
			
			GlobalsSetValue("new_enemies_global_no_recoil", "0")
		end
		
		local components = EntityGetComponent( mwyah, "VariableStorageComponent" )
		if ( components ~= nil ) then
			for _,comp_id in pairs(components) do 
				local var_name = ComponentGetValue( comp_id, "name" )
				if ( var_name == "state") then
					local mwyah_state = ComponentGetValue2(comp_id, "value_int")
					if mwyah_state < 6 then 
						GlobalsSetValue("new_enemies_global_no_recoil", "0")
						
						function split_string(inputstr, sep)
						  sep = sep or "%s"
						  local t= {}
						  for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
							table.insert(t, str)
						  end
						  return t
						end

						function toboolean(str)
							return str == "1"
						end

						local character_platforming_component = EntityGetFirstComponentIncludingDisabled(entity_id, "CharacterPlatformingComponent")
						local var_store = get_variable_storage_component(entity_id, "new_enemies_player_mwyah_phase2_old_values")
						if ( var_store ~= nil ) and ( var_store > 0 ) then
							local stored_values = ComponentGetValue2(var_store, "value_string")
							local properties = split_string(stored_values, ",")
							
							for i, v in ipairs(properties) do
								local name, prop_type, value = string.match(v, "(.+):(.+)=(.+)")
								value = _G["to"..prop_type](value)
								ComponentSetValue2(character_platforming_component, name, value)
							end
							
							EntityRemoveComponent(entity_id, var_store)
							
						end
						remove_status_effects( entity_id )
						EntityRemoveComponent(entity_id, GetUpdatedComponentID())
					end
				end
			end
		end
	elseif ( mwyah2 ~= nil ) and  ( mwyah2 > 0 ) then
		
		local children = EntityGetAllChildren( mwyah2, "pitcheck_b" )
		local arena_center_id = children[1]
		
		local var_comp_laser = get_variable_storage_component( arena_center_id, "is_laser" )
		local current_state = ComponentGetValue2(var_comp_laser, "value_int")
		
		if current_state == 1 then
			if ( EntityHasTag( projectile_id, "projectile_not" ) == false ) then
				local new_proj = shoot_projectile_from_projectile(projectile_id, "data/entities/projectiles/deck/teleport_mwyah_invisible_check.xml", px, py, vx, vy)
				EntityAddChild(entity_id, new_proj)
			end
			
			local all_comps = EntityGetAllComponents( projectile_id ) or {}
			for i, comp in ipairs(all_comps) do
				EntityRemoveComponent( projectile_id, comp )
			end
			
			EntityApplyTransform(projectile_id, 66666666, 66666666)
			
			local all_comps = EntityGetAllComponents( projectile_id ) or {}
			for i, comp in ipairs(all_comps) do
				EntityRemoveComponent( projectile_id, comp )
			end
			
			local lua_comp = EntityAddComponent2(projectile_id, "LuaComponent", {
				script_source_file="data/scripts/animals/kill_entity.lua",
				execute_every_n_frame=1,
				execute_on_added=false
			})
			
			GlobalsSetValue("new_enemies_global_no_recoil", "1")
			add_status_effect( entity_id, "data/ui_gfx/status_indicators/telebolt_effect.png", "$telebolts_status_name", "$telebolts_status_desc" )
		else
			GlobalsSetValue("new_enemies_global_no_recoil", "0")
			remove_status_effects( entity_id )
		end
		
		function split_string(inputstr, sep)
		  sep = sep or "%s"
		  local t= {}
		  for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
			table.insert(t, str)
		  end
		  return t
		end

		function toboolean(str)
			return str == "1"
		end

		local character_platforming_component = EntityGetFirstComponentIncludingDisabled(entity_id, "CharacterPlatformingComponent")
		local var_store = get_variable_storage_component(entity_id, "new_enemies_player_mwyah_phase2_old_values")
		if ( var_store ~= nil ) and ( var_store > 0 ) then
			local stored_values = ComponentGetValue2(var_store, "value_string")
			local properties = split_string(stored_values, ",")
			
			for i, v in ipairs(properties) do
				local name, prop_type, value = string.match(v, "(.+):(.+)=(.+)")
				value = _G["to"..prop_type](value)
				ComponentSetValue2(character_platforming_component, name, value)
			end
			
			EntityRemoveComponent(entity_id, var_store)
			
		end
	else
		GlobalsSetValue("new_enemies_global_no_recoil", "0")
		
		function split_string(inputstr, sep)
		  sep = sep or "%s"
		  local t= {}
		  for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
			table.insert(t, str)
		  end
		  return t
		end

		function toboolean(str)
			return str == "1"
		end

		local character_platforming_component = EntityGetFirstComponentIncludingDisabled(entity_id, "CharacterPlatformingComponent")
		local var_store = get_variable_storage_component(entity_id, "new_enemies_player_mwyah_phase2_old_values")
		if ( var_store ~= nil ) and ( var_store > 0 ) then
			local stored_values = ComponentGetValue2(var_store, "value_string")
			local properties = split_string(stored_values, ",")
			
			for i, v in ipairs(properties) do
				local name, prop_type, value = string.match(v, "(.+):(.+)=(.+)")
				value = _G["to"..prop_type](value)
				ComponentSetValue2(character_platforming_component, name, value)
			end
			
			EntityRemoveComponent(entity_id, var_store)
			
		end
		EntityRemoveComponent(entity_id, this_lua)
		remove_status_effects( entity_id )
	end
end