dofile_once("data/scripts/lib/utilities.lua")
dofile_once("data/scripts/gun/procedural/gun_action_utils.lua")

local entity_id = GetUpdatedEntityID()
local x, y, dir, sx, sy = EntityGetTransform(entity_id)

function getDistance(x1, y1, x2, y2)
    return math.sqrt(math.pow((y2 - y1), 2) + math.pow((x2 - x1), 2))
end

local sprite_width = 9
local max_distance = 600

-- Calculate the reflection direction
function reflect(x, y, normal_x, normal_y)
    local dot = x * normal_x + y * normal_y
    return x - 2 * dot * normal_x, y - 2 * dot * normal_y
end

function handleLaserReflection(x, y, dir, proc_val)
    local reflections = {}
    local current_x, current_y, current_dir = x, y, dir
    local remaining_proc_val = proc_val

    while remaining_proc_val >= 0 do
        local x2 = current_x + math.cos(current_dir) * 288
        local y2 = current_y + math.sin(current_dir) * 288
        local did_hit, hit_x, hit_y = RaytraceSurfaces(current_x, current_y, x2, y2)

        if did_hit then
            table.insert(reflections, {x = current_x, y = current_y, dir = current_dir, hit_x = hit_x, hit_y = hit_y})
            local distance = getDistance(current_x, current_y, hit_x, hit_y)
            local scale = math.min(distance, max_distance) / sprite_width
            EntityLoad("data/entities/particles/particle_explosion/eradicator_laser_explosion.xml", hit_x, hit_y)

            -- Get the surface normal at the collision point
            local found_normal, normal_x, normal_y, approximate_distance = GetSurfaceNormal(hit_x, hit_y, 10, 10)
            if not found_normal then break end

            -- Reflect the direction
            local new_dir_x, new_dir_y = reflect(math.cos(current_dir), math.sin(current_dir), normal_x, normal_y)
            current_dir = math.atan2(new_dir_y, new_dir_x)

            current_x, current_y = hit_x, hit_y
            remaining_proc_val = remaining_proc_val - 1
        else
            table.insert(reflections, {x = current_x, y = current_y, dir = current_dir, hit_x = current_x + math.cos(current_dir) * max_distance, hit_y = current_y + math.sin(current_dir) * max_distance})
            break
        end
    end

    return reflections
end

local components = EntityGetComponent(entity_id, "VariableStorageComponent") or {}
local proc_val = 0

for _, comp_id in pairs(components) do
    local var_name = ComponentGetValue(comp_id, "name")
    if var_name == "proc_val" then
        proc_val = ComponentGetValue2(comp_id, "value_int")
    end
end

local reflections = handleLaserReflection(x, y, dir, proc_val)

for _, reflection in ipairs(reflections) do
    local distance = getDistance(reflection.x, reflection.y, reflection.hit_x, reflection.hit_y)
    local scale = math.min(distance, max_distance) / sprite_width
    EntitySetTransform(entity_id, reflection.x, reflection.y, reflection.dir, scale, sy)

    if proc_val > 0 then
        local new_projectile_id = shoot_projectile_from_projectile(entity_id, "data/entities/projectiles/mechahub_laser_spawner_proc.xml", reflection.hit_x, reflection.hit_y, math.cos(reflection.dir), math.sin(reflection.dir))
        EntitySetTransform(new_projectile_id, reflection.hit_x, reflection.hit_y, reflection.dir, 1, 1)

        local new_components = EntityGetComponent(new_projectile_id, "VariableStorageComponent") or {}
        for _, new_comp_id in pairs(new_components) do
            local new_var_name = ComponentGetValue(new_comp_id, "name")
            if new_var_name == "proc_val" then
                ComponentSetValue2(new_comp_id, "value_int", proc_val - 1)
            end
        end
    end
end

-- Set enemy's x velocity to 0
function set_controls_enabled(id, enabled)
    local entity = id
    if entity then
        local controls_component = EntityGetFirstComponentIncludingDisabled(entity, "ControlsComponent")
        ComponentSetValue2(controls_component, "enabled", enabled)
    end
end

local parent_id = EntityGetParent(entity_id)

set_controls_enabled(parent_id, false)

local character_data_component = EntityGetFirstComponentIncludingDisabled(parent_id, "CharacterDataComponent")

if ComponentGetValue2(character_data_component, "is_on_ground") == true or ComponentGetValue2(character_data_component, "is_on_slippery_ground") == true then
    local velocity_component = EntityGetFirstComponentIncludingDisabled(parent_id, "VelocityComponent")
    local vel_x, vel_y = ComponentGetValue2(velocity_component, "mVelocity")

    ComponentSetValue2(character_data_component, "mVelocity", 0, vel_y)
    ComponentSetValue2(velocity_component, "mVelocity", 0, vel_y)
    ComponentSetValue2(character_data_component, "ground_stickyness", 1)
end

local sprite_components = EntityGetComponent(entity_id, "SpriteComponent") or {}
for _, sprite_component in ipairs(sprite_components) do
    local image = ComponentGetValue2(sprite_component, "image_file")
    if proc_val > 0 and image == "data/projectiles_gfx/eradicator_laser.png" then
        ComponentSetValue2(sprite_component, "image_file", "data/projectiles_gfx/eradicator_laser_full.png")
    elseif proc_val <= 0 and image == "data/projectiles_gfx/eradicator_laser_full.png" then
        ComponentSetValue2(sprite_component, "image_file", "data/projectiles_gfx/eradicator_laser.png")
    end
end