dofile_once("data/scripts/lib/utilities.lua")

function math.round(num)
    return math.floor(num + 0.5)
end

-- Calculate the reflection direction
function reflect(x_ref, y_ref, normal_x, normal_y)
    -- Calculate the angle of the surface normal
    local normal_angle = math.atan2(normal_y, normal_x)
    
    -- Round the angle to the nearest multiple of 90 degrees
    local rounded_angle = math.round(normal_angle / (math.pi / 2)) * (math.pi / 2)
    
    -- Get the rounded surface normal components
    local rounded_normal_x = math.cos(rounded_angle)
    local rounded_normal_y = math.sin(rounded_angle)

    -- Reflect the direction
    local dot = x_ref * rounded_normal_x + y_ref * rounded_normal_y
    return x_ref - 2 * dot * rounded_normal_x, y_ref - 2 * dot * rounded_normal_y
end

local entity_id = GetUpdatedEntityID()
local x, y, dir, sx, sy = EntityGetTransform(entity_id)

local offset_y = 0

if y > 0 then
	offset_y = 9999999
end

y = y - offset_y

function getDistance(x1, y1, x2, y2)
    return math.sqrt(math.pow((y2 - y1), 2) + math.pow((x2 - x1), 2))
end

function math.findIntersect(l1p1x, l1p1y, l1p2x, l1p2y, l2p1x, l2p1y, l2p2x, l2p2y, seg1, seg2)
    local a1, b1, a2, b2 = l1p2y - l1p1y, l1p1x - l1p2x, l2p2y - l2p1y, l2p1x - l2p2x
    local c1, c2 = a1 * l1p1x + b1 * l1p1y, a2 * l2p1x + b2 * l2p1y
    local det = a1 * b2 - a2 * b1
    if det == 0 then return false, "The lines are parallel." end
    local x_int, y_int = (b2 * c1 - b1 * c2) / det, (a1 * c2 - a2 * c1) / det
    if seg1 or seg2 then
        local min, max = math.min, math.max
        if seg1 and not (min(l1p1x, l1p2x) <= x_int and x_int <= max(l1p1x, l1p2x) and min(l1p1y, l1p2y) <= y_int and y_int <= max(l1p1y, l1p2y)) or
           seg2 and not (min(l2p1x, l2p2x) <= x_int and x_int <= max(l2p1x, l2p2x) and min(l2p1y, l2p2y) <= y_int and y_int <= max(l2p1y, l2p2y)) then
            return false, "The lines don't intersect."
        end
    end
    return x_int, y_int
end

-- local test_intersectx, test_intersecty = math.findIntersect(- 999999, 100, 999999, 100, cx - wall_offset, y - 999999, cx - wall_offset, y + 999999)
-- if test_intersectx and test_intersecty then
	-- GamePrint("Yes this intersects")
-- end

local root_id = EntityGetRootEntity(entity_id)
local wall_offset = 200
local max_depth = 41400

if (root_id ~= nil) and (root_id > 0) then
	local components = EntityGetComponent( root_id, "VariableStorageComponent" )
	if ( components ~= nil ) then
		for _,comp_id in pairs(components) do 
			local var_name = ComponentGetValue( comp_id, "name" )
			if ( var_name == "wall_offset") then
				wall_offset = ComponentGetValue2(comp_id, "value_float")
			end
			if ( var_name == "max_depth") then
				max_depth = ComponentGetValue2(comp_id, "value_float")
			end
		end
	end
end

local cx, cy = 0, 0
for _, child in ipairs(EntityGetAllChildren(root_id) or {}) do
    if EntityGetName(child) == "sides" then
        cx, cy = EntityGetTransform(child)
    end
end

local components = EntityGetComponentIncludingDisabled(entity_id, "VariableStorageComponent") or {}
local proc_val = 0
local length = 600
for _, comp_id in pairs(components) do
    local var_name = ComponentGetValue(comp_id, "name")
    if var_name == "proc_val" then
        proc_val = ComponentGetValue2(comp_id, "value_int")
    end
    if var_name == "length" then
		length_var_comp = comp_id
        length = ComponentGetValue2(comp_id, "value_float")
		length = length
    end
end


local sprite_width = 9
local max_distance = 600

local x2 = x + math.cos(dir) * length
local y2 = y + math.sin(dir) * length

local original_x, original_y = EntityGetTransform(entity_id)

local x3 = original_x + math.cos(dir) * max_distance
local y3 = original_y + math.sin(dir) * max_distance

local did_hit, hit_x, hit_y = RaytraceSurfaces(original_x, original_y, x3, y3)

local intersect_x1_first, intersect_y1_first = math.findIntersect(x, y, x2, y2, cx - wall_offset, y - 9999999999, cx - wall_offset, y + 9999999999)
local intersect_x2_first, intersect_y2_first = math.findIntersect(x, y, x2, y2, cx + wall_offset, y - 9999999999, cx + wall_offset, y + 9999999999)

if did_hit and ( ( hit_x >= cx - wall_offset ) and ( hit_x <= cx + wall_offset ) ) and ( original_y <= max_depth ) then
	local distance = getDistance(original_x, original_y, hit_x, hit_y)
	local scale = math.min(distance, length) / sprite_width
	ComponentSetValue2(length_var_comp, "value_float", distance)
	EntitySetTransform(entity_id, original_x, original_y, dir, scale, sy)
elseif (intersect_x1_first and intersect_y1_first and x >= cx - wall_offset and x2 < cx - wall_offset) and ( original_y <= max_depth ) then
	   
	if intersect_x1_first and intersect_y1_first then
		local distance = getDistance(x, y, intersect_x1_first, intersect_y1_first)
		local scale = math.min(distance, length) / sprite_width
		ComponentSetValue2(length_var_comp, "value_float", distance)
		EntitySetTransform(entity_id, original_x, original_y, dir, scale, sy)
	end
elseif (intersect_x2_first and intersect_y2_first and x <= cx + wall_offset and x2 > cx + wall_offset) and ( original_y <= max_depth ) then
	if intersect_x2_first and intersect_y2_first then
		local distance = getDistance(x, y, intersect_x2_first, intersect_y2_first)
		local scale = math.min(distance, length) / sprite_width
		ComponentSetValue2(length_var_comp, "value_float", distance)
		EntitySetTransform(entity_id, original_x, original_y, dir, scale, sy)
	end
elseif ( x < cx - wall_offset ) or ( x > cx + wall_offset ) or ( original_y > max_depth ) then
	distance = 0

	local scale = math.min(distance, max_distance) / sprite_width

	EntitySetTransform(entity_id, original_x, original_y, dir , scale, sy)
else
	local distance = 600
	local scale = math.min(distance, length) / sprite_width
	-- ComponentSetValue2(length_var_comp, "value_float", distance)
	EntitySetTransform(entity_id, original_x, original_y, dir, scale, sy)
	local spritecomponents = EntityGetComponentIncludingDisabled(entity_id, "SpriteComponent") or {}
	for _, comp_id in pairs(spritecomponents) do
		local filename = ComponentGetValue(comp_id, "image_file")
		if filename == "data/projectiles_gfx/eradicator_laser_full.png" then
			ComponentSetValue2(comp_id, "image_file", "data/projectiles_gfx/eradicator_laser_half.png")
		end
	end
end

function handleLaserReflection(x, y, dir, proc_val)
    local reflections = {}
    local current_x_original, current_y_original, current_dir = original_x, original_y, dir
    local remaining_proc_val = proc_val

    while remaining_proc_val > 0 do
		local current_x, current_y = current_x_original, current_y_original
		if current_y > 0 then
			current_y = current_y - offset_y
		end
	
        local x2 = current_x + math.cos(current_dir) * max_distance
        local y2 = current_y + math.sin(current_dir) * max_distance
		
        local x3 = current_x_original + math.cos(current_dir) * max_distance
        local y3 = current_y_original + math.sin(current_dir) * max_distance
		
		-- GamePrint("current y is: " .. tostring(current_y) )
		
        -- Check for intersection with imaginary walls
        local intersect_x1, intersect_y1 = math.findIntersect(current_x, current_y, x2, y2, cx - wall_offset, y - 9999999999, cx - wall_offset, y + 9999999999)
        local intersect_x2, intersect_y2 = math.findIntersect(current_x, current_y, x2, y2, cx + wall_offset, y - 9999999999, cx + wall_offset, y + 9999999999)
        local did_hit2, hit_x2, hit_y2 = RaytraceSurfaces(current_x_original, current_y_original, x3, y3)
		if did_hit2 and ( ( hit_x2 >= cx - wall_offset ) and ( hit_x2 <= cx + wall_offset ) ) and ( current_y_original <= max_depth ) then
            -- Laser hits a surface within wall boundaries
            local distance = getDistance(current_x_original, current_y_original, hit_x2, hit_y2)
            local scale = math.min(distance, max_distance) / sprite_width
			
			EntityLoad("data/entities/particles/particle_explosion/eradicator_laser_explosion.xml", hit_x2, hit_y2)

            -- Get the surface normal at the collision point
            local found_normal, normal_x, normal_y, approximate_distance = GetSurfaceNormal(hit_x2, hit_y2, 16, 10)
            if not found_normal then 
				break 
			end

            -- Reflect the direction
            local new_dir_x, new_dir_y = reflect(math.cos(current_dir), math.sin(current_dir), normal_x, normal_y)
            current_dir = math.atan2(new_dir_y, new_dir_x)

            current_x_original, current_y_original = hit_x2, hit_y2
            remaining_proc_val = remaining_proc_val - 1

            -- Spawn the laser spawner at the reflection point
            local velocity_x = math.cos(current_dir)
            local velocity_y = math.sin(current_dir)
            local proj_filename = "data/entities/projectiles/mechahub_laser_spawner_proc.xml"
            if remaining_proc_val == 0 then
                proj_filename = "data/entities/projectiles/mechahub_laser_spawner_proc_last.xml"
            end
			local new_projectile_id = shoot_projectile_from_projectile(entity_id, proj_filename, hit_x2, hit_y2, velocity_x, velocity_y)
            EntitySetTransform(new_projectile_id, hit_x2, hit_y2)
			if (root_id ~= nil) and (root_id > 0) then
				EntityAddChild(root_id, new_projectile_id)
			end
			local varcomponents = EntityGetComponentIncludingDisabled(new_projectile_id, "VariableStorageComponent") or {}
			for _, comp_id in pairs(varcomponents) do
				local var_name = ComponentGetValue(comp_id, "name")
				if var_name == "length" then
					ComponentSetValue2(comp_id, "value_float", distance)
				end
			end
        elseif (intersect_x1 and intersect_y1 and current_x_original >= cx - wall_offset and x2 < cx - wall_offset) and ( current_y_original <= max_depth ) then
			
			-- local laser_comps = EntityGetComponentIncludingDisabled(entity_id, "LaserEmitterComponent") or {}
			-- if ( #laser_comps > 0 ) then
				-- for i, comp in ipairs(laser_comps) do
					-- ComponentObjectSetValue2( comp, "laser", "max_length", distance )
				-- end
			-- end
			
            -- Laser hits imaginary wall
            local hit_x, hit_y
            local normal_x, normal_y
            if intersect_x1 and intersect_y1 then
                hit_x, hit_y = intersect_x1, intersect_y1 + offset_y
                normal_x, normal_y = -1, 0
				-- GamePrint("Hit left imaginary wall " .. tostring(hit_x) .. " " .. tostring(hit_y))
				EntityLoad("data/entities/particles/particle_explosion/eradicator_laser_explosion.xml", hit_x, hit_y)
            end
			
			local distance = getDistance(current_x_original, current_y_original, hit_x, hit_y)
			-- GamePrint("Current Distance: " .. tostring(distance))
            -- Reflect the direction
            local new_dir_x, new_dir_y = reflect(math.cos(current_dir), math.sin(current_dir), normal_x, normal_y)
            current_dir = math.atan2(new_dir_y, new_dir_x)

            current_x_original, current_y_original = hit_x, hit_y
            remaining_proc_val = remaining_proc_val - 1

            -- Spawn the laser spawner at the reflection point
            local velocity_x = math.cos(current_dir)
            local velocity_y = math.sin(current_dir)
            local proj_filename = "data/entities/projectiles/mechahub_laser_spawner_proc.xml"
            if remaining_proc_val == 0 then
                proj_filename = "data/entities/projectiles/mechahub_laser_spawner_proc_last.xml"
            end
			local new_projectile_id = shoot_projectile_from_projectile(entity_id, proj_filename, hit_x2, hit_y2, velocity_x, velocity_y)
            EntitySetTransform(new_projectile_id, hit_x, hit_y)
			if (root_id ~= nil) and (root_id > 0) then
				EntityAddChild(root_id, new_projectile_id)
			end
			
			local varcomponents = EntityGetComponentIncludingDisabled(new_projectile_id, "VariableStorageComponent") or {}
			for _, comp_id in pairs(varcomponents) do
				local var_name = ComponentGetValue(comp_id, "name")
				if var_name == "length" then
					ComponentSetValue2(comp_id, "value_float", distance)
				end
			end
        elseif (intersect_x2 and intersect_y2 and current_x_original <= cx + wall_offset and x2 > cx + wall_offset) and ( current_y_original <= max_depth ) then
			
            -- Laser hits imaginary wall
            local hit_x, hit_y
            local normal_x, normal_y
			if intersect_x2 and intersect_y2 then
                hit_x, hit_y = intersect_x2, intersect_y2 + offset_y
                normal_x, normal_y = 1, 0
				-- GamePrint("Hit right imaginary wall")
				EntityLoad("data/entities/particles/particle_explosion/eradicator_laser_explosion.xml", hit_x, hit_y)
            end
			
			local distance = getDistance(current_x_original, current_y_original, hit_x, hit_y)
			-- GamePrint("Current Distance: " .. tostring(distance))

            -- Reflect the direction
            local new_dir_x, new_dir_y = reflect(math.cos(current_dir), math.sin(current_dir), normal_x, normal_y)
            current_dir = math.atan2(new_dir_y, new_dir_x)

            current_x_original, current_y_original = hit_x, hit_y
            remaining_proc_val = remaining_proc_val - 1

            -- Spawn the laser spawner at the reflection point
            local velocity_x = math.cos(current_dir)
            local velocity_y = math.sin(current_dir)
            local proj_filename = "data/entities/projectiles/mechahub_laser_spawner_proc.xml"
            if remaining_proc_val == 0 then
                proj_filename = "data/entities/projectiles/mechahub_laser_spawner_proc_last.xml"
            end
			local new_projectile_id = shoot_projectile_from_projectile(entity_id, proj_filename, hit_x2, hit_y2, velocity_x, velocity_y)
            EntitySetTransform(new_projectile_id, hit_x, hit_y)
			
			if (root_id ~= nil) and (root_id > 0) then
				EntityAddChild(root_id, new_projectile_id)
			end
			local varcomponents = EntityGetComponentIncludingDisabled(new_projectile_id, "VariableStorageComponent") or {}
			for _, comp_id in pairs(varcomponents) do
				local var_name = ComponentGetValue(comp_id, "name")
				if var_name == "length" then
					ComponentSetValue2(comp_id, "value_float", distance)
				end
			end
		else
            break
        end
    end
end

handleLaserReflection(x, y, dir, proc_val)