dofile_once("data/scripts/lib/utilities.lua")

function math.round(num)
    return math.floor(num + 0.5)
end

function getDistance(x1, y1, x2, y2)
    return math.sqrt(math.pow((y2 - y1), 2) + math.pow((x2 - x1), 2))
end

-- Calculate the reflection direction
function reflect(x_ref, y_ref, normal_x, normal_y)
    -- Calculate the angle of the surface normal
    local normal_angle = math.atan2(normal_y, normal_x)
    
    -- Round the angle to the nearest multiple of 90 degrees
    local rounded_angle = math.round(normal_angle / (math.pi / 2)) * (math.pi / 2)
    
    -- Get the rounded surface normal components
    local rounded_normal_x = math.cos(rounded_angle)
    local rounded_normal_y = math.sin(rounded_angle)

    -- Reflect the direction
    local dot = x_ref * rounded_normal_x + y_ref * rounded_normal_y
    return x_ref - 2 * dot * rounded_normal_x, y_ref - 2 * dot * rounded_normal_y
end

local sprite_width = 9
local max_distance = 600

local entity_id = GetUpdatedEntityID()
local x, y, dir, sx, sy = EntityGetTransform(entity_id)

local offset_y = 0

if y > 0 then
	offset_y = 9999999
end

local root_id = EntityGetRootEntity(entity_id)
local wall_offset = 200
local max_depth = 41400

if (root_id ~= nil) and (root_id > 0) then
	local components = EntityGetComponent( root_id, "VariableStorageComponent" )
	if ( components ~= nil ) then
		for _,comp_id in pairs(components) do 
			local var_name = ComponentGetValue( comp_id, "name" )
			if ( var_name == "wall_offset") then
				wall_offset = ComponentGetValue2(comp_id, "value_float")
			end
			if ( var_name == "max_depth") then
				max_depth = ComponentGetValue2(comp_id, "value_float")
			end
		end
	end
end

local cx, cy = 0, 0
for _, child in ipairs(EntityGetAllChildren(root_id) or {}) do
    if EntityGetName(child) == "sides" then
        cx, cy = EntityGetTransform(child)
    end
end

local x3 = x + math.cos(dir) * max_distance
local y3 = y + math.sin(dir) * max_distance

-- GameCreateSpriteForXFrames( "data/debug/sui_close_button.png", x, y, true, 0, 0, 1, true )


local did_hit, hit_x, hit_y = RaytraceSurfaces(x, y, x3, y3)

--and ( ( hit_x >= cx - wall_offset ) and ( hit_x <= cx + wall_offset ) )
if did_hit and ( hit_x <= cx + wall_offset ) and ( hit_x >= cx - wall_offset ) then

	local distance = getDistance(x, y, hit_x, hit_y)
	local scale = math.min(distance, max_distance) / sprite_width
	-- ComponentSetValue2(length_var_comp, "value_float", distance)
	EntitySetTransform(entity_id, x, y, dir, scale, sy)
	return
end

y = y - offset_y

function math.findIntersect(l1p1x, l1p1y, l1p2x, l1p2y, l2p1x, l2p1y, l2p2x, l2p2y, seg1, seg2)
    local a1, b1, a2, b2 = l1p2y - l1p1y, l1p1x - l1p2x, l2p2y - l2p1y, l2p1x - l2p2x
    local c1, c2 = a1 * l1p1x + b1 * l1p1y, a2 * l2p1x + b2 * l2p1y
    local det = a1 * b2 - a2 * b1
    if det == 0 then return false, "The lines are parallel." end
    local x_int, y_int = (b2 * c1 - b1 * c2) / det, (a1 * c2 - a2 * c1) / det
    if seg1 or seg2 then
        local min, max = math.min, math.max
        if seg1 and not (min(l1p1x, l1p2x) <= x_int and x_int <= max(l1p1x, l1p2x) and min(l1p1y, l1p2y) <= y_int and y_int <= max(l1p1y, l1p2y)) or
           seg2 and not (min(l2p1x, l2p2x) <= x_int and x_int <= max(l2p1x, l2p2x) and min(l2p1y, l2p2y) <= y_int and y_int <= max(l2p1y, l2p2y)) then
            return false, "The lines don't intersect."
        end
    end
    return x_int, y_int
end

-- local test_intersectx, test_intersecty = math.findIntersect(- 999999, 100, 999999, 100, cx - wall_offset, y - 999999, cx - wall_offset, y + 999999)
-- if test_intersectx and test_intersecty then
	-- GamePrint("Yes this intersects")
-- end



local components = EntityGetComponentIncludingDisabled(entity_id, "VariableStorageComponent") or {}
local proc_val = 0
local length = 600
for _, comp_id in pairs(components) do
    local var_name = ComponentGetValue(comp_id, "name")
    if var_name == "length" then
		length_var_comp = comp_id
        length = ComponentGetValue2(comp_id, "value_float")
		length = length
		-- GamePrint(length)
    end
end

local x2 = x + math.cos(dir) * max_distance
local y2 = y + math.sin(dir) * max_distance

local intersect_x1_first, intersect_y1_first = math.findIntersect(x, y, x2, y2, cx - wall_offset, y - 9999999999, cx - wall_offset, y + 9999999999)
local intersect_x2_first, intersect_y2_first = math.findIntersect(x, y, x2, y2, cx + wall_offset, y - 9999999999, cx + wall_offset, y + 9999999999)

local original_x, original_y = EntityGetTransform(entity_id)

if (intersect_x1_first and intersect_y1_first and x >= cx - wall_offset and x2 < cx - wall_offset) and ( original_y <= max_depth ) then
	   
	if intersect_x1_first and intersect_y1_first then
		local distance = getDistance(x, y, intersect_x1_first, intersect_y1_first)
		local scale = math.min(distance, max_distance) / sprite_width
		-- ComponentSetValue2(length_var_comp, "value_float", distance)
		EntitySetTransform(entity_id, original_x, original_y, dir, scale, sy)
	end
elseif (intersect_x2_first and intersect_y2_first and x <= cx + wall_offset and x2 > cx + wall_offset) and ( original_y <= max_depth ) then
	if intersect_x2_first and intersect_y2_first then
		local distance = getDistance(x, y, intersect_x2_first, intersect_y2_first)
		local scale = math.min(distance, max_distance) / sprite_width
		-- ComponentSetValue2(length_var_comp, "value_float", distance)
		EntitySetTransform(entity_id, original_x, original_y, dir, scale, sy)
	end
elseif ( x < cx - wall_offset ) or ( x > cx + wall_offset ) or ( original_y > max_depth ) then
	distance = 0

	local scale = math.min(distance, max_distance) / sprite_width

	EntitySetTransform(entity_id, original_x, original_y, dir , scale, sy)
else
	distance= max_distance

	local scale = math.min(distance, max_distance) / sprite_width

	EntitySetTransform(entity_id, original_x, original_y, dir , scale, sy)
	-- local spritecomponents = EntityGetComponentIncludingDisabled(entity_id, "SpriteComponent") or {}
	-- for _, comp_id in pairs(spritecomponents) do
		-- local filename = ComponentGetValue(comp_id, "image_file")
		-- if filename == "data/projectiles_gfx/eradicator_laser_full.png" then
			-- ComponentSetValue2(comp_id, "image_file", "data/projectiles_gfx/eradicator_laser_half.png")
		-- end
	-- end
end